#
# Copyright (C) 2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np


def seed_list(seed, nseeds):
    rng = np.random.Generator(np.random.MT19937(seed=seed))
    used = {}
    seeds = []
    for _ in range(nseeds):
        candidate = rng.integers(0, np.iinfo(np.uint32).max)
        while candidate in used:
            candidate = rng.integers(0, np.iinfo(np.uint32).max)
        used[candidate] = 1
        seeds.append(candidate)
    return seeds



import unittest

import fwdpy11
import gsl_error


class testGSLerror(unittest.TestCase):
    def testErrorHandler(self):
        with self.assertRaises(RuntimeError):
            gsl_error.trigger_error_handle_locally()
        try:
            gsl_error.trigger_error_handle_locally()
        except RuntimeError as e:
            self.assertEqual(str(e), "matrix not square")

    def testUnhandledGSLError(self):
        with self.assertRaises(fwdpy11.GSLError):
            gsl_error.trigger_error()


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import typing

import fwdpy11
from fwdpy11.conditional_models import (ConditionalModelOutput,
                                        NewMutationParameters,
                                        SimulationStatus)

from ._track_added_mutation import _track_added_mutation

__disallowed_kwargs = ["until"]


# NOTE: this entire work flow can impose NASTY side-effects:
#  * We are taking lots of references to s.asdict()tuff in the input params.
#  * In general, the input values can be modified by the simulation.
#  * This includes demographic models, genetic value stuff, etc.
#  * Ideally, we'd subjet it all to a copy.deepcopy, and we should test that...
def _selective_sweep(
    rng: fwdpy11.GSLrng,
    pop: fwdpy11._types.DiploidPopulation,
    params: fwdpy11.ModelParams,
    mutation_parameters: NewMutationParameters,
    stopping_condition: typing.Callable[
        [fwdpy11.DiploidPopulation, int, typing.Tuple[float, float, int]],
        SimulationStatus,
    ],
    **kwargs,
) -> ConditionalModelOutput:
    """
    This function is a wrapper around
    :func:`fwdpy11.conditional_models.track_added_mutation`.

    This function requires a `stopping_condition`.
    If `when` is not given as a keyword argument, it is assumed to be 0.
    The keyword argument `until` is not allowed.
    """
    if "when" not in kwargs or kwargs["when"] is None:
        kwargs["when"] = 0

    for d in __disallowed_kwargs:
        if d in kwargs:
            raise ValueError(f"{d} is not a valid kwarg")

    return _track_added_mutation(
        rng,
        pop,
        params,
        mutation_parameters,
        stopping_condition=stopping_condition,
        **kwargs,
    )



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import typing

import attr

import fwdpy11._fwdpy11

from .class_decorators import (attr_add_asblack, attr_class_pickle_with_super,
                               attr_class_to_from_dict)


def _is_integer_if_discrete(self, attribute, value):
    if self.discrete is True:
        attr.validators.instance_of(int)(self, attribute, value)


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class PoissonInterval(fwdpy11._fwdpy11._ll_PoissonInterval):
    """
    Generate poisson number of crossover breakpoints.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: The beginning of the region
    :type beg: int or float
    :param end: The end of the region
    :type end: int or float
    :param mean: The mean number of breakpoints per meiosis
    :type mean: float
    :param discrete: If `False`, positions are continuous
                     and uniform from `[beg, end)`.
                     If `True`, positions take integer values
                     uniformly from `[beg, end)`.
    :type discrete: bool

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.5.0

        Refactored back-end to be based on fwdpp types

    .. versionchanged:: 0.7.1

        Add __repr__

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class

    .. versionchanged:: 0.12.0

        Added `discrete` option to initializer.
    """

    beg: typing.Union[int, float] = attr.ib(validator=_is_integer_if_discrete)
    end: typing.Union[int, float] = attr.ib(validator=_is_integer_if_discrete)
    mean: float
    discrete: bool = attr.ib(kw_only=True, default=False)

    def __attrs_post_init__(self):
        super(PoissonInterval, self).__init__(
            beg=self.beg, end=self.end, mean=self.mean, discrete=self.discrete
        )


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class PoissonPoint(fwdpy11._fwdpy11._ll_PoissonPoint):
    """
    Generate a recombination breakpoint at a fixed position if the
    number of crossover events is odd.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param position: The position of the crossover
    :type position: int or float
    :param mean: The mean number of breakpoints per meiosis
    :type mean: float
    :param discrete: If `False`, positions are continuous and
                     uniform from `[beg, end)`. If `True`,
                     positions take integer values uniformly
                     from `[beg, end)`.
    :type discrete: bool

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.5.0

        Refactored back-end to be based on fwdpp types

    .. versionchanged:: 0.7.1

        Add __repr__

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class

    .. versionchanged:: 0.12.0

        Added `discrete` option to initializer.
    """

    position: typing.Union[int, float] = \
        attr.ib(validator=_is_integer_if_discrete)
    mean: float
    discrete: bool = attr.ib(kw_only=True, default=False)

    def __attrs_post_init__(self):
        super(PoissonPoint, self).__init__(
            position=self.position, mean=self.mean, discrete=self.discrete
        )

    def __getstate__(self):
        return self.asdict()

    def __setstate__(self, d):
        self.__dict__.update(d)
        super(PoissonPoint, self).__init__(**d)


@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class BinomialInterval(fwdpy11._fwdpy11._ll_BinomialInterval):
    """
    Generate exactly one crossover with a given probability

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: The beginning of the region
    :type beg: int or float
    :param end: The end of the region
    :type end: int or float
    :param probability: The probability of a recombination (per meiosis).
    :type probability: float
    :param discrete: If `False`, positions are continuous
                     and uniform from `[beg, end)`.
                     If `True`, positions take integer values
                     uniformly from `[beg, end)`.
    :type discrete: bool

    .. versionadded:: 0.5.2

    .. versionchanged:: 0.7.1

        Add __repr__

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class

    .. versionchanged:: 0.12.0

        Added `discrete` option to initializer.
    """

    beg: typing.Union[int, float] = attr.ib(validator=_is_integer_if_discrete)
    end: typing.Union[int, float] = attr.ib(validator=_is_integer_if_discrete)
    probability: float
    discrete: bool = attr.ib(kw_only=True, default=False)

    def __attrs_post_init__(self):
        super(BinomialInterval, self).__init__(
            beg=self.beg,
            end=self.end,
            probability=self.probability,
            discrete=self.discrete,
        )


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class BinomialPoint(fwdpy11._fwdpy11._ll_BinomialPoint):
    """
    Generate a crossover breakpoint at a fixed position with a
    fixed probability.  This class represents genetic distance
    as centiMorgans/100.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param position: The beginning of the region
    :type position: int or float
    :param probability: The probability of a recombination (per meiosis).
    :type probability: float
    :param discrete: If `False`, positions are continuous
                     and uniform from `[beg, end)`.
                     If `True`, positions take integer values
                     uniformly from `[beg, end)`.
    :type discrete: bool

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.5.0

        Refactored back-end to be based on fwdpp types

    .. versionchanged:: 0.7.1

            Add __repr__

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class

    .. versionchanged:: 0.12.0

        Added `discrete` option to initializer.
    """

    position: typing.Union[int, float] = attr.ib(
        validator=_is_integer_if_discrete)
    probability: typing.Union[int, float]
    discrete: bool = attr.ib(kw_only=True, default=False)

    def __attrs_post_init__(self):
        super(BinomialPoint, self).__init__(
            position=self.position, probability=self.probability,
            discrete=self.discrete
        )


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class FixedCrossovers(fwdpy11._fwdpy11._ll_FixedCrossovers):
    """
    Generate a fixed number of crossover breakpoints.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: The beginning of the region
    :type beg: int or float
    :param end: The end of the region
    :type end: int or float
    :param num_xovers: The number of breakpoints per meiosis
    :type num_xovers: float
    :param discrete: If `False`, positions are continuous
                     and uniform from `[beg, end)`.
                     If `True`, positions take integer values
                     uniformly from `[beg, end)`.
    :type discrete: bool

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.5.0

        Refactored back-end to be based on fwdpp types

    .. versionchanged:: 0.7.1

        Add __repr__

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class

    .. versionchanged:: 0.12.0

        Added `discrete` option to initializer.
    """

    beg: typing.Union[int, float] = attr.ib(validator=_is_integer_if_discrete)
    end: typing.Union[int, float] = attr.ib(validator=_is_integer_if_discrete)
    num_xovers: int
    discrete: bool = attr.ib(kw_only=True, default=False)

    def __attrs_post_init__(self):
        super(FixedCrossovers, self).__init__(
            beg=self.beg,
            end=self.end,
            num_xovers=self.num_xovers,
            discrete=self.discrete,
        )


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class BinomialIntervalMap(fwdpy11._fwdpy11._ll_BinomialIntervalMap):
    """
    Generate exactly one crossover with a given probability
    and a non-uniform map to generate positions.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param probability: The probability of a breakpoint
    :type probability: float
    :param regions: the map describing intervals and weights
    :type regions: list[fwdpy11.Region]
    :param discrete: If `False`, positions are continuous
                     and uniform from `[beg, end)`.
                     If `True`, positions take integer values
                     uniformly from `[beg, end)`.
    :type discrete: bool

    Unlike :class:`fwdpy11.BinomialInterval`, this class allows
    a non-uniform map to generate breakpoint positions.
    The ``regions`` parameter is used to generate a lookup
    table allowing efficient generation of positions.

    .. versionadded:: 0.20.0
    """
    probability: float
    regions: typing.List[fwdpy11.Region]
    discrete: bool = attr.ib(kw_only=True, default=False)

    def __attrs_post_init__(self):
        super(BinomialIntervalMap, self).__init__(
            probability=self.probability,
            regions=self.regions,
            discrete=self.discrete
        )



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import typing

import attr
import tskit  # type: ignore

from .._fwdpy11 import Mutation
from ._flags import (INDIVIDUAL_IS_ALIVE, INDIVIDUAL_IS_FIRST_GENERATION,
                     INDIVIDUAL_IS_PRESERVED)


@attr.s(auto_attribs=True, frozen=True, kw_only=True, eq=True)
class DiploidMetadata(object):
    """
    Instances of this type are returned by
    :func:`fwdpy11.tskit_tools.decode_individual_metadata`.

    The following ``kwargs`` are used to construct instances of
    this type. Member attributes have the same names as these
    ``kwargs```:

    :param g: Genetic value
    :type g: float
    :param e: "Random/environmental" value
    :type e: float
    :param w: Fitness
    :type w: float
    :param sex: Sex
    :type sex: int
    :param deme: Deme
    :type deme: int
    :param label: Index in the population at time of living/sampling
    :type label: int
    :param alive: Individual is alive
    :type alive: bool
    :param preserved: Individual is preserved
    :type preserved: bool
    :param first_generation: Individual is first generation
    :type first_generation: bool
    :param parents: ``label`` of parents
    :type parents: list
    :param geography: Location in 3d space
    :type geography: list
    :param nodes: Nodes on tree sequence
    :type nodes: list

    .. versionadded:: 0.12.0
    """

    g: float
    e: float
    w: float
    sex: int
    deme: int
    label: int
    alive: bool = attr.ib(converter=bool)
    preserved: bool = attr.ib(converter=bool)
    first_generation: bool = attr.ib(converter=bool)
    parents: typing.List[int]
    geography: typing.List[float]
    nodes: typing.List[int]

    @classmethod
    def from_table_row(cls, ind: tskit.IndividualTableRow):
        """
        Generate a metadata object from a :class:`tskit.IndividualTableRow`

        .. versionadded:: 0.15.0
        """
        alive = ind.flags & INDIVIDUAL_IS_ALIVE
        preserved = ind.flags & INDIVIDUAL_IS_PRESERVED
        first_generation = ind.flags & INDIVIDUAL_IS_FIRST_GENERATION
        return cls(
            **ind.metadata,
            alive=alive,
            preserved=preserved,
            first_generation=first_generation
        )


def decode_individual_metadata(
    ts: tskit.TreeSequence,
    rows=None
) -> typing.List[DiploidMetadata]:
    """
    Decodes a :class:`tskit.IndividualTable`.

    :param ts: A tree sequence
    :type ts: :class:`tskit.TreeSequence`

    :param rows: The rows to decode.
                 If `None`, the entire table is decoded and returned.

    :returns: individual metadata
    :rtype: List of :class:`fwdpy11.tskit_tools.DiploidMetadata`

    .. warning::

        Decoding the entire table may consume a lot of system memory.

    .. versionadded:: 0.12.0

    .. versionchanged:: 0.15.0

        Add index/slice access to the table.

    .. versionchanged:: 0.17.0

        Change input from TableCollection to TreeSequence
    """
    rv = []

    try:
        ind = ts.individual(rows)
        rv.append(DiploidMetadata.from_table_row(ind))
    except Exception:
        if rows is None:
            rows = range(0, ts.num_individuals, 1)
        elif isinstance(rows, slice):
            rows = range(rows.start, rows.stop, rows.step)
        for i in rows:
            ind = ts.individual(i)
            rv.append(DiploidMetadata.from_table_row(ind))

    return rv


def _append_mutation_metadata(ts, site, md, mutations):
    if md is not None:
        if "esizes" not in md:
            mutations.append(
                Mutation(
                    pos=ts.site(site).position,
                    s=md["s"],
                    h=md["h"],
                    g=md["origin"],
                    label=md["label"],
                )
            )
        else:
            mutations.append(
                Mutation(
                    pos=ts.site(site).position,
                    s=md["s"],
                    h=md["h"],
                    g=md["origin"],
                    esizes=md["esizes"],
                    heffects=md["heffects"],
                    label=md["label"],
                )
            )
    else:
        mutations.append(None)


def decode_mutation_metadata(
    ts: tskit.TreeSequence,
    rows=None
) -> typing.List[typing.Optional[Mutation]]:
    """
    Decodes metadata from a :class:`tskit.MutationTable`.

    :param ts: A tree sequence
    :type ts: :class:`tskit.TreeSequence`
    :param rows: The rows to decode.
                 If `None`, the entire table is decoded and returned.

    :returns: Mutations
    :rtype: list

    .. warning::

        Decoding the entire table may consume a lot of system memory.

    .. versionadded:: 0.12.0

    .. versionchanged:: 0.15.0

        Return type is now a list, allowing
        some elements to be `None`.

        Add index/slice access to the table.

    .. versionchanged:: 0.17.0

        Change input from TableCollection to TreeSequence
    """
    mutations: typing.List = []
    try:
        m = ts.mutation(rows)
        _append_mutation_metadata(ts, m.site, m.metadata, mutations)
    except Exception:
        if rows is None:
            rows = range(0, ts.num_mutations, 1)
        elif isinstance(rows, slice):
            rows = range(rows.start, rows.stop, rows.step)
        for i in rows:
            m = ts.mutation(i)
            _append_mutation_metadata(ts, m.site, m.metadata, mutations)
    return mutations



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import demes
import fwdpy11
import numpy as np
import pytest


@pytest.fixture(scope="function")
def numpy_generator(request):
    try:
        seed = request.param.get("seed", None)
        return np.random.Generator(np.random.MT19937(seed=seed))
    except AttributeError as a:  # NOQA
        raise a


@pytest.fixture(scope="function")
def rng(request):
    try:
        seed = request.param.get("seed", None)
        return fwdpy11.GSLrng(seed)
    except AttributeError:  # NOQA
        return fwdpy11.GSLrng(42)


@pytest.fixture(scope="function")
def pop(request):
    popsize = request.param["N"]
    genome_length = request.param["genome_length"]
    return fwdpy11.DiploidPopulation(popsize, genome_length)


@pytest.fixture(scope="function")
def mslike_pop(request):
    try:
        N = request.param.get("N", None)
        return fwdpy11.DiploidPopulation(N, 1.0)
    except AttributeError:  # NOQA
        return fwdpy11.DiploidPopulation(1000, 1.0)


@pytest.fixture
def gutenkunst():
    yaml = """
description: The Gutenkunst et al. (2009) OOA model.
doi:
- https://doi.org/10.1371/journal.pgen.1000695
time_units: years
generation_time: 25

demes:
- name: ancestral
  description: Equilibrium/root population
  epochs:
  - {end_time: 220e3, start_size: 7300}
- name: AMH
  description: Anatomically modern humans
  ancestors: [ancestral]
  epochs:
  - {end_time: 140e3, start_size: 12300}
- name: OOA
  description: Bottleneck out-of-Africa population
  ancestors: [AMH]
  epochs:
  - {end_time: 21.2e3, start_size: 2100}
- name: YRI
  description: Yoruba in Ibadan, Nigeria
  ancestors: [AMH]
  epochs:
  - start_size: 12300
- name: CEU
  description: Utah Residents (CEPH) with Northern and Western European Ancestry
  ancestors: [OOA]
  epochs:
  - {start_size: 1000, end_size: 29725}
- name: CHB
  description: Han Chinese in Beijing, China
  ancestors: [OOA]
  epochs:
  - {start_size: 510, end_size: 54090}

migrations:
- {demes: [YRI, OOA], rate: 25e-5}
- {demes: [YRI, CEU], rate: 3e-5}
- {demes: [YRI, CHB], rate: 1.9e-5}
- {demes: [CEU, CHB], rate: 9.6e-5}
"""
    return demes.loads(yaml)


@pytest.fixture
def small_split_model():
    yaml = """
description: Two deme model.
time_units: generations
demes:
- name: ancestral
  description: ancestral deme
  epochs:
  - {start_size: 100}
- name: deme1
  description: child deme
  start_time: 100
  epochs:
  - {start_size: 100}
  ancestors: [ancestral]
"""
    return demes.loads(yaml)


@pytest.fixture
def two_deme_split_with_ancestral_size_change():
    yaml = """
description: Two deme model with migration and size changes.
time_units: generations
demes:
- name: ancestral
  description: ancestral deme, two epochs
  epochs:
  - {end_time: 20, start_size: 100}
  - {end_time: 10, start_size: 200}
- name: deme1
  description: child 1
  epochs:
  - {start_size: 250, end_size: 500, end_time: 0}
  ancestors: [ancestral]
- name: deme2
  description: child 2
  epochs:
  - {start_size: 50, end_size: 200, end_time: 0}
  ancestors: [ancestral]
migrations:
- {demes: [deme1, deme2], rate: 1e-3}
"""
    return demes.loads(yaml)


@pytest.fixture
def start_demes_model_with_two_pops():
    yaml = """
description: after slicing original graph at time 50 (below slice)
time_units: generations
demes:
- name: deme1
  epochs:
  - {start_size: 100, end_time: 50}
  - {start_size: 100, end_size: 200, end_time: 0}
- name: deme2
  epochs:
  - {start_size: 75, end_time: 50}
  - {start_size: 75, end_time: 20}
  - {start_size: 300, end_time: 0}
migrations:
- demes: [deme1, deme2]
  rate: 1e-2
  start_time: 50
"""
    return demes.loads(yaml)



from .data_matrix_from_tables import (data_matrix_from_tables,  # NOQA
                                      make_data_matrix)
from .import_demes import demography_from_demes  # NOQa
from .simplify_tables import simplify, simplify_tables  # NOQA

from fwdpy11._fwdpy11 import _infinite_sites
from fwdpy11 import GSLrng, DiploidPopulation


def infinite_sites(rng: GSLrng, pop: DiploidPopulation, mu: float) -> int:
    return _infinite_sites(rng, pop, mu)


def _validate_regions(regions, sequence_length):
    for r in regions:
        try:
            if r.beg >= sequence_length or r.end > sequence_length:
                raise ValueError(
                    f"Region {r} extends beyond the"
                    + f" sequence length {sequence_length}")
        except AttributeError:
            try:
                for des in r.des:
                    if des.beg >= sequence_length or des.end > sequence_length:
                        raise ValueError(
                            f"Region {des} in {r} extends beyond" +
                            f" the sequence length {sequence_length}")
            except TypeError:  # r.des not Iterable
                if r.des.beg >= sequence_length or r.des.end > sequence_length:
                    raise ValueError(
                        f"Region {des} in {r} extends beyond the" +
                        f" sequence length {sequence_length}")



import unittest

import fwdpy11
import numpy as np


class test_stopping_criterion_DiploidPopulation(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(1000, 1.0)
        p = {
            "nregions": [],  # No neutral mutations -- add them later!
            "gvalue": fwdpy11.Additive(2.0),
            "sregions": [fwdpy11.ExpS(0, 1, 1, -0.1)],
            "recregions": [fwdpy11.PoissonInterval(0, 1, 1e-3)],
            "rates": (0.0, 1e-3, None),
            # Keep mutations at frequency 1 in the pop if they affect fitness.
            "prune_selected": False,
            "demography": fwdpy11.ForwardDemesGraph.tubes(self.pop.deme_sizes()[1],
                                                          burnin=10*self.pop.N,
                                                          burnin_is_exact=True),
            "simlen": 10 * self.pop.N,
        }
        self.params = fwdpy11.ModelParams(**p)

    def test_stop(self):
        rng = fwdpy11.GSLrng(42)

        def generation(pop, simplified):
            if pop.generation >= 50:
                return True
            return False

        fwdpy11.evolvets(rng, self.pop, self.params, 100,
                         stopping_criterion=generation)
        self.assertEqual(self.pop.generation, 50)


def test_stopping_criterion_with_ancient_samples():
    """
    This triggers github issue 844
    """

    def preserve(pop, sampler):
        if pop.generation == 10:
            sampler.assign(np.arange(pop.N, dtype=np.uint32))

    def stopper(pop, _):
        if pop.generation == 10:
            return True
        return False

    pop = fwdpy11.DiploidPopulation(100, 1.0)
    pdict = {
        "sregions": [fwdpy11.ExpS(0, 1, 1, -0.05, 1)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0.1, 0),
        "simlen": 20,
        "demography": fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                      burnin=20,
                                                      burnin_is_exact=True)
    }
    params = fwdpy11.ModelParams(**pdict)
    rng = fwdpy11.GSLrng(90210)

    fwdpy11.evolvets(
        rng, pop, params, 100, recorder=preserve, stopping_criterion=stopper
    )
    assert pop.generation == 10


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

import fwdpy11
import msprime
import numpy as np


class TestConversion(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.ts = msprime.simulate(10, recombination_rate=0.025, Ne=1000)

    # def testGetTablesDiscretizeTime(self):
    #     n, e, ntips, l = fwdpy11.ts_from_msprime._convert_tables(self.ts)
    #     self.assertEqual(ntips, 10)
    #     self.assertEqual(l, 1.0)
    #     na = np.array(n, copy=False)
    #     tzero = np.where(na['time'] == 0.0)
    #     self.assertTrue(len(tzero[0]) == 10)

    def testCreateDiploidPopulation(self):
        pop = fwdpy11.DiploidPopulation.create_from_tskit(self.ts)
        self.assertEqual(pop.N, 5)
        self.assertEqual(pop.tables.genome_length, 1.0)
        md = np.array(pop.diploid_metadata, copy=False)
        n = md["nodes"].flatten()
        self.assertTrue(np.array_equal(n, np.arange(2 * pop.N, dtype=n.dtype)))


class TestConversionFromMultipleDemes(unittest.TestCase):
    def test_deme_field_of_metadata(self):
        nodes_per_deme = 500
        Ne = 3 * nodes_per_deme // 2
        Nr = 50.0
        config = [
            msprime.PopulationConfiguration(nodes_per_deme),
            msprime.PopulationConfiguration(nodes_per_deme),
            msprime.PopulationConfiguration(nodes_per_deme),
        ]

        events = [
            msprime.MassMigration(1 * Ne, 1, 0, 1.0),
            msprime.MassMigration(1.5 * Ne, 2, 0, 1.0),
        ]

        ts = msprime.simulate(
            population_configurations=config,
            Ne=Ne,
            random_seed=98765,
            recombination_rate=Nr / Ne,
            demographic_events=events,
        )
        pop = fwdpy11.DiploidPopulation.create_from_tskit(ts)
        self.assertEqual(pop.N, 750)
        alive_nodes = pop.alive_nodes
        self.assertEqual(len(alive_nodes), 2 * pop.N)
        for i in range(3):
            for j in alive_nodes[i * nodes_per_deme: (i + 1) * nodes_per_deme]:
                self.assertEqual(pop.tables.nodes[j].deme, i)
            k, el = i * nodes_per_deme // 2, (i + 1) * nodes_per_deme // 2
            for j in pop.diploid_metadata[k:el]:
                self.assertEqual(j.deme, i)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017-2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import typing

from hypothesis import settings, given
from hypothesis.strategies import integers

import attr
import fwdpy11.tskit_tools
import fwdpy11.tskit_tools._dump_tables_to_tskit
import numpy as np
import pytest
import tskit

np.random.seed(54321)

MOCK_NODE_TABLE_DTYPE = [("time", np.float64), ("deme", np.int32)]
MOCK_EDGE_TABLE_DTYPE = [
    ("left", np.float64),
    ("right", np.float64),
    ("parent", np.int32),
    ("child", np.int32),
]


@attr.s(auto_attribs=True, frozen=True, kw_only=True, eq=True)
class MockDiploid(object):
    g: float
    e: float
    w: float
    sex: int
    deme: int
    label: int
    parents: typing.List[int]
    geography: typing.List[float]
    nodes: typing.List[int]


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class MockMutationRecord(object):
    node: int
    key: int
    site: int
    derived_state: int


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class MockSite(object):
    position: float
    ancestral_state: int


def empty_mock_node_table() -> np.ndarray:
    return np.array([], dtype=MOCK_NODE_TABLE_DTYPE)


def empty_mock_edge_table() -> np.ndarray:
    return np.array([], dtype=MOCK_EDGE_TABLE_DTYPE)


@attr.s(auto_attribs=True)
class MockTables(object):
    genome_length: float = 1.0
    nodes: np.ndarray = empty_mock_node_table()
    edges: np.ndarray = empty_mock_edge_table()
    mutations: typing.List[MockMutationRecord] = attr.Factory(list)
    sites: typing.List[MockSite] = attr.Factory(list)


@attr.s(auto_attribs=True, kw_only=True)
class MockPopulation(object):
    diploid_metadata: typing.List[MockDiploid] = attr.Factory(list)
    ancient_sample_metadata: typing.List[MockDiploid] = attr.Factory(list)
    mutations: fwdpy11.MutationVector = attr.ib()
    tables: MockTables = attr.ib()

    @mutations.default
    def _init_mutations(self):
        return fwdpy11.MutationVector()

    @tables.default
    def _init_tables(self):
        return MockTables()


def add_diploid_metadata_and_nodes(pop: MockPopulation) -> MockPopulation:
    # Add diploids
    demes = [0, 1, 1, 2]
    node_table_data = []
    for i in range(2):
        node_table_data.append((1, demes[i]))
        node_table_data.append((1, demes[i]))
        pop.diploid_metadata.append(
            MockDiploid(
                g=float(i),
                e=float(i ** 2),
                w=float(i ** 2),
                sex=i,
                label=i,
                deme=demes[i],
                parents=[101, 202],  # nonsense data
                geography=[i, 2 * i, 3 * i],  # nonsense data
                nodes=[2 * i, 2 * i + 1],
            )
        )

    # Ancient samples, also generation 0 == first generation preserved.
    for i in range(2, 4):
        node_table_data.append((0, demes[i]))
        node_table_data.append((0, demes[i]))
        pop.ancient_sample_metadata.append(
            MockDiploid(
                g=float(i),
                e=float(i ** 2),
                w=float(i ** 2),
                sex=i,
                label=i,
                deme=demes[i],
                parents=[101, 202],  # nonsense data
                geography=[i, 2 * i, 3 * i],  # nonsense data
                nodes=[2 * i, 2 * i + 1],
            )
        )
    pop.tables.nodes = np.array(node_table_data, dtype=MOCK_NODE_TABLE_DTYPE)

    return pop


def add_mutation(pop: MockPopulation, node: int, with_vectors: bool):
    pass


@pytest.fixture
def mock_population() -> MockPopulation:
    pop = MockPopulation()

    return pop


@pytest.fixture
def tables() -> tskit.TableCollection:
    tc = tskit.TableCollection(1.0)
    return tc


def test_diploid_metadata(mock_population, tables):
    mock_population = add_diploid_metadata_and_nodes(mock_population)

    fwdpy11.tskit_tools._dump_tables_to_tskit._initializeIndividualTable(
        mock_population, tables
    )

    ts = tables.tree_sequence()

    decoded = fwdpy11.tskit_tools.decode_individual_metadata(ts)

    for i in decoded[: len(mock_population.diploid_metadata)]:
        assert i.alive
        assert not i.preserved
        assert not i.first_generation

    for i in decoded[len(mock_population.diploid_metadata):]:
        assert not i.alive
        assert i.preserved
        assert i.first_generation

    for i, j in zip(
        decoded,
        mock_population.diploid_metadata + mock_population.ancient_sample_metadata,
    ):
        assert i.g == j.g
        assert i.e == j.e
        assert i.w == j.w
        assert i.sex == j.sex
        assert i.deme == j.deme
        assert i.label == j.label
        assert i.geography == j.geography
        assert i.nodes == j.nodes


def test_mutation_metadata_without_vectors(mock_population, tables):
    mock_population.generation = 1

    mock_population.tables.sites.append(
        MockSite(position=0.1, ancestral_state=0))

    mock_population.tables.mutations.append(
        MockMutationRecord(node=0, key=0, site=0, derived_state=1)
    )

    mock_population.mutations.append(
        fwdpy11.Mutation(pos=0.1, s=-1e-3, h=1,
                         g=mock_population.generation, label=7)
    )

    fwdpy11.tskit_tools._dump_tables_to_tskit._dump_mutation_site_and_site_tables(
        mock_population, tables
    )

    tables.nodes.add_row(time=0.0)
    ts = tables.tree_sequence()
    decoded = fwdpy11.tskit_tools.decode_mutation_metadata(ts)

    assert len(decoded) == len(mock_population.tables.mutations)

    for i, j in zip(decoded, mock_population.mutations):
        assert i.pos == j.pos
        assert i.s == j.s
        assert i.h == j.h
        assert i.g == j.g
        assert i.neutral == j.neutral
        assert i.label == j.label
        assert len(i.esizes) == 0
        assert len(i.heffects) == 0


def test_mutation_metadata_with_vectors(mock_population, tables):
    mock_population.generation = 1

    mock_population.tables.sites.append(
        MockSite(position=0.1, ancestral_state=0))

    mock_population.tables.mutations.append(
        MockMutationRecord(node=0, key=0, site=0, derived_state=1)
    )

    mock_population.mutations.append(
        fwdpy11.Mutation(
            pos=0.1,
            s=-1e-3,
            h=1,
            g=mock_population.generation,
            label=7,
            esizes=np.arange(4).tolist(),
            heffects=np.array([1.0] * 4).tolist(),
        )
    )

    assert len(mock_population.mutations[0].esizes) == 4

    fwdpy11.tskit_tools._dump_tables_to_tskit._dump_mutation_site_and_site_tables(
        mock_population, tables
    )

    tables.nodes.add_row(time=0.0)
    ts = tables.tree_sequence()
    decoded = fwdpy11.tskit_tools.decode_mutation_metadata(ts)

    assert len(decoded) == len(mock_population.tables.mutations)

    for i, j in zip(decoded, mock_population.mutations):
        assert i.pos == j.pos
        assert i.s == j.s
        assert i.h == j.h
        assert i.g == j.g
        assert i.neutral == j.neutral
        assert i.label == j.label
        assert len(i.esizes) == 4
        assert len(i.heffects) == 4
        assert np.array_equal(i.esizes, j.esizes)
        assert np.array_equal(i.heffects, j.heffects)


@given(msprime_seed=integers(1, int(2**32 - 1)),
       fp11_seed=integers(1, int(2**32-1)))
@settings(deadline=None, max_examples=5)
def test_encoding_neutral_mutations_with_msprime_tree_sequences(
    msprime_seed, fp11_seed
):
    """
    This is a test for GitHub issue 656.
    This test is part of release 0.13.0
    Previous versions of fwdpy11 used uint32_t for Mutation.g.
    The unsigned integer caused overflow issues when adding mutations
    to trees with negative node times.
    Mutation.g was changes to int32_t in 0.13.0 as part of a fix.
    """
    import fwdpy11
    import msprime

    ts = msprime.simulate(500, Ne=1000, random_seed=msprime_seed)
    pop = fwdpy11.DiploidPopulation.create_from_tskit(ts)
    rng = fwdpy11.GSLrng(fp11_seed)
    _ = fwdpy11.infinite_sites(rng, pop, 1e-3)
    for m in pop.tables.mutations:
        assert pop.mutations[m.key].g <= pop.tables.nodes[m.node].time
    _ = pop.dump_tables_to_tskit()



"""
Test that all the examples are valid instances of the schema.

Run with ``python3 -m pytest ``
"""
import pathlib
import functools

import pytest
import jsonschema
from ruamel.yaml import YAML
import hypothesis
import hypothesis_jsonschema


@functools.lru_cache(maxsize=None)
def load_yaml(filename):
    with open(filename, encoding="utf-8") as f:
        with YAML(typ="safe") as yaml:
            data = yaml.load(f)
    return data


def validate(yaml_path):
    schema = load_yaml("schema/hdm-v1.0.yaml")
    data = load_yaml(yaml_path)
    jsonschema.validate(instance=data, schema=schema)


@pytest.mark.parametrize(
    "yaml_path", map(str, pathlib.Path("examples/").glob("**/*.yaml"))
)
def test_examples(yaml_path):
    validate(yaml_path)


@pytest.mark.parametrize(
    "yaml_path", map(str, pathlib.Path("test-cases/valid").glob("**/*.yaml"))
)
def test_test_cases_valid(yaml_path):
    validate(yaml_path)


# Check that the MDM schema is a subschema of the HDM schema by generating
# random data matching the MDM schema and validating against the HDM schema.
@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.too_slow])
@hypothesis.given(
    mdm_data=hypothesis_jsonschema.from_schema(load_yaml("schema/mdm-v1.0.yaml"))
)
def test_subschema(mdm_data):
    hdm_schema = load_yaml("schema/hdm-v1.0.yaml")
    jsonschema.validate(instance=mdm_data, schema=hdm_schema)



from typing import Iterable

import numpy as np
import scipy.sparse  # type: ignore

from .._fwdpy11 import Edge, MutationRecord, Node, Site, ll_TableCollection

# Functions related to FS calculation

NOT_A_SAMPLE = np.iinfo(np.int32).min


def _include_both(m):
    return True


def _include_neutral(m):
    return m.neutral is True


def _include_selected(m):
    return m.neutral is False


def _validate_windows(windows, genome_length):
    if windows != sorted(windows, key=lambda x: x[0]):
        raise ValueError("windows must be sorted in increasing order")
    for w in windows:
        if w[0] >= w[1]:
            raise ValueError("windows must be [a, b) and a < b")
        for i in w:
            if i < 0 or i > genome_length:
                raise ValueError(
                    "window coordinates must be [0, genome_length)")
    for i in range(1, len(windows), 2):
        if windows[i - 1][0] < windows[i][1] and \
                windows[i][0] < windows[i - 1][1]:
            raise ValueError("windows cannot overlap")


def _handle_fs_marginalizing(fs, marginalize, nwindows, num_sample_groups):
    if marginalize is False or num_sample_groups == 1:
        return fs
    if nwindows == 1:
        temp = {}
        for i in range(num_sample_groups):
            axes = num_sample_groups - i - 1
            temp[i] = np.ma.array(np.asarray(fs.sum(axis=axes).flatten())[0])
            temp[i][0] = np.ma.masked
            temp[i][-1] = np.ma.masked
        return temp

    rv = []
    for f in fs:
        temp = {}
        for i in range(num_sample_groups):
            axes = num_sample_groups - i - 1
            temp[i] = np.ma.array(np.asarray(f.sum(axis=axes).flatten())[0])
            temp[i][0] = np.ma.masked
            temp[i][-1] = np.ma.masked
        rv.append(temp)
    return rv


def _update_window(windex, tree, windows):
    while windex < len(windows) and tree.right > windows[windex][1]:
        windex += 1
    return windex


def _position_in_window(pos, window):
    return pos >= window[0] and pos < window[1]


def _simplify(tables, samples, simplify):
    import fwdpy11
    if simplify is False:
        return tables, samples

    t, i = fwdpy11.simplify_tables(tables, samples)
    return t, i[samples]


class TableCollection(ll_TableCollection):
    def __init__(self, *args):
        super(TableCollection, self).__init__(*args)

    @property
    def L(self) -> float:
        return self._genome_length

    @property
    def genome_length(self) -> float:
        """Genome length"""
        return self._genome_length

    @property
    def edges(self) -> Iterable[Edge]:
        """Supports buffer protocol"""
        return self._edges

    @property
    def nodes(self) -> Iterable[Node]:
        """Supports buffer protocol"""
        return self._nodes

    @property
    def sites(self) -> Iterable[Site]:
        """Supports buffer protocol"""
        return self._sites

    @property
    def mutations(self) -> Iterable[MutationRecord]:
        """Supports buffer protocol"""
        return self._mutations

    @property
    def input_left(self):
        """Edge input order"""
        return self._input_left

    @property
    def output_right(self):
        """Edge output order"""
        return self._output_right

    def build_indexes(self):
        """
        Build edge input/output indexes.

        .. versionadded:: 0.13.0

        """
        self._build_indexes()

    def _1dfs(self, samples, windows, include_function, simplify):
        """
        Returns an array with the zero and fixed
        bins masked out.  The masking is for
        consistency w/ndfs output.
        """
        from . import TreeIterator

        t, s = _simplify(self, samples, simplify)
        fs = [np.zeros(len(s) + 1, dtype=np.int32) for i in windows]
        ti = TreeIterator(t, s)
        windex = 0
        sites = np.array(t.sites, copy=False)
        positions = sites["position"][:]
        for tree in ti:
            for m in tree.mutations():
                if include_function(m):
                    pos = positions[m.site]
                    while windex < len(windows) and windows[windex][1] < pos:
                        windex += 1
                    if windex >= len(windows):
                        break
                    if _position_in_window(pos, windows[windex]):
                        c = np.uint32(tree.leaf_counts(m.node))
                        fs[windex][c] += 1

        rv = []
        for i in fs:
            i = np.ma.array(i)
            i[0] = np.ma.masked
            i[-1] = np.ma.masked
            rv.append(i)
        return rv

    def _ndfs(
        self,
        samples,
        sample_groups,
        num_sample_groups,
        windows,
        include_function,
        simplify,
    ):
        """
        For more than one sample, always work with the joint FS
        at first. When the marginals are wanted, we extract them
        later.
        """
        from . import TreeIterator

        shapes = tuple(len(i) + 1 for i in samples)
        dok_JFS = [scipy.sparse.dok_matrix(
            shapes, dtype=np.int32) for i in windows]

        sample_list = np.where(sample_groups != NOT_A_SAMPLE)[0]
        t, s = _simplify(self, sample_list, simplify)
        ti = TreeIterator(t, s, update_samples=True)
        counts = np.zeros(len(samples), dtype=np.int32)
        windex = 0
        sites = np.array(t.sites, copy=False)
        positions = sites["position"][:]
        for tree in ti:
            for m in tree.mutations():
                if include_function(m):
                    pos = positions[m.site]
                    while windex < len(windows) and windows[windex][1] < pos:
                        windex += 1
                    if windex >= len(windows):
                        break
                    if _position_in_window(pos, windows[windex]):
                        counts[:] = 0
                        d = tree.samples_below(m.node)
                        if len(d) > 0:
                            c = np.unique(sample_groups[d], return_counts=True)
                            counts[c[0]] += c[1]
                            dok_JFS[windex][tuple((i) for i in counts)] += 1
        return [scipy.sparse.coo_matrix(i) for i in dok_JFS]

    def _fs_implementation(self, samples, windows, include_function, simplify):
        """
        self is a fwdpy11.TableCollection
        samples is a list of 1d numpy.ndarray
        """
        if len(samples) == 0:
            raise ValueError("empty list of samples")

        # Don't figure out sample groups when
        # there is only 1
        if len(samples) == 1:
            return self._1dfs(samples[0], windows, include_function, simplify)

        sample_groups = np.array(
            [NOT_A_SAMPLE] * len(self.nodes), dtype=np.int32)
        for i, j in enumerate(samples):
            if np.any(sample_groups[j] != NOT_A_SAMPLE):
                raise ValueError(
                    "sample nodes cannot be part of multiple sample groups"
                )
            sample_groups[j] = i

        num_sample_groups = i + 1
        return self._ndfs(
            samples,
            sample_groups,
            num_sample_groups,
            windows,
            include_function,
            simplify,
        )

    def fs(
        self,
        samples,
        marginalize=False,
        simplify=False,
        windows=None,
        separate_windows=False,
        include_neutral=True,
        include_selected=True,
    ):
        """
        :param samples: lists of numpy arrays of sample nodes
        :param marginalize: For ``FS`` involving multiple samples,
                            extract out the marginal ``FS`` per sample.
        :param simplify: If ``True``, simplify with respect to the sample
                         set prior to calculating the ``FS``.
        :param windows: A list of non-overlapping intervals from which
                        the ``FS`` is calculated.
        :param separate_windows: If ``True``, return ``FS`` separately for
                                 each interval in ``windows``.
        :param include_neutral: Include neutral mutations?
        :param include_selected: Include selected mutations?

        :returns: The mutation frequency spectrum.  The `dtype` is `int32`.
        :rtype: object

        The details of the return value depend heavily on the options.

        * If a single sample list is provided, the return value is a
          1-d ndarray with the first and last bins masked.  The unmasked
          bins correspond to frequencies 1 to n-1, where n is the number
          of sample nodes.
        * If multiple sample lists are provided, the return value is a
          :class:`scipy.sparse.coo_matrix`.  For each dimension, the 0 and n
          bins are included.
        * If ``marginalize == True`` and more than one sample bin is provided,
          the :class:`scipy.sparse.coo_matrix` matrix is converted into a dict
          where the key is the index of the sample list and the
          value is a dense 1-d array,
          masked as described above.
        * If multiple windows are provided and ``separate_windows == False``,
          then the return value is a single frequency spectrum summed over
          windows.
          If ``separate_windows == True``, then a list of frequency spectra is
          returned, indexed in the same order as the input windows.

        .. versionadded:: 0.6.0

            Python implementation added

        .. versionchanged:: 0.18.0

            Dropped `sparse` as a dependency, using types from `scipy.sparse`
            instead. This change drops support for more than 2 sample lists.

        """
        for s in samples:
            if len(s) == 0:
                raise ValueError("samples lists cannot be empty")
            if len(s) < 2:
                raise ValueError("samples lists must have at least two nodes")
            if np.any(s >= len(self.nodes)):
                raise ValueError("invalid samples")

        if windows is None:
            windows = [(0, self.genome_length)]

        _validate_windows(windows, self.genome_length)

        if include_neutral is False and include_selected is False:
            raise ValueError(
                "One or both of include_neutral "
                "and include_selected must be True"
            )

        include_function = _include_both
        if include_neutral is False:
            include_function = _include_selected
        elif include_selected is False:
            include_function = _include_neutral

        lfs = self._fs_implementation(
            samples, windows, include_function, simplify)
        if separate_windows is True:
            lfs = _handle_fs_marginalizing(
                lfs, marginalize, len(windows), len(samples))
            return lfs
        fs = lfs[0]
        for i in lfs[1:]:
            fs += i
        fs = _handle_fs_marginalizing(
            fs, marginalize, len(windows), len(samples))
        return fs

    def preserved_nodes(self):
        raise AttributeError(
            "TableCollection.preserved_nodes was removed in 0.13.0 and is"
            " now an attribute of DiploidPopulation"
        )



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import typing
import warnings

import attr
import fwdpy11
import numpy as np
from fwdpy11.class_decorators import (attr_add_asblack,
                                      attr_class_to_from_dict_no_recurse)

from fwdpy11._types.demographic_model_details import DemographicModelDetails
from fwdpy11._types.forward_demes_graph import ForwardDemesGraph


@attr_add_asblack
@attr_class_to_from_dict_no_recurse
@attr.s(kw_only=True, frozen=True, slots=True, repr_ns="fwdpy11")
class MutationAndRecombinationRates(object):
    """
    Stores and validates the mutation and recombination rate parameters
    of a simulation.

    Instances of this class are created by ``kwargs`` that populate
    attributes of the same name:

    :param neutral_mutation_rate:
    :type neutral_mutation_rate: float
    :param selected_mutation_rate:
    :type selected_mutation_rate: float
    :param recombination_rate:
    :type recombination_rate: float or None

    Instances of this class are passed as the ``rates`` ``kwarg``
    to :class:`fwdpy11.ModelParams`.

    .. versionadded:: 0.8.0
    """

    neutral_mutation_rate: float = attr.ib(converter=float)
    selected_mutation_rate: float = attr.ib(converter=float)
    recombination_rate = attr.ib()

    @neutral_mutation_rate.validator
    @selected_mutation_rate.validator
    def validate_individual_rate(self, attribute, value):
        if not np.isfinite(value):
            raise ValueError(
                f"{attribute} must be finite, but we got {value} instead")
        if value < 0.0:
            raise ValueError(
                f"{attribute} must be >= 0.0, but we got {value} instead")

    @recombination_rate.validator
    def validate_recombination_rate(self, attribute, value):
        if value is None:
            return
        try:
            value = float(value)
        except TypeError:
            raise ValueError(
                f"{attribute} must be convertible to float, but we got {value} instead"
            )
        attr.validators.instance_of(float)(self, attribute, value)
        self.validate_individual_rate(attribute, value)


def _convert_rates(value):
    if isinstance(value, MutationAndRecombinationRates):
        return value

    try:
        return MutationAndRecombinationRates(
            neutral_mutation_rate=value[0],
            selected_mutation_rate=value[1],
            recombination_rate=value[2],
        )
    except KeyError:
        return MutationAndRecombinationRates(**value)


@attr_add_asblack
@attr_class_to_from_dict_no_recurse
@attr.s(kw_only=True, frozen=True, slots=True, repr_ns="fwdpy11")
class ModelParams(object):
    """
    This class stores and validates the parameters of a simulation.
    Instances of this class are constructed using ``kwargs``
    and instance attributes are immutable once initialized.

    This class accepts the following ``kwargs``, which are also
    the attribute names for instances:

    :param nregions: List of regions where neutral mutations occur
    :type nregions: list[fwdpy11.Region]
    :param sregions: List of regions where selected mutations occur
    :type sregions: list[fwdpy11.Sregion]
    :param recregions: List of regions where recombination events occur
    :type recregions: list[fwdpy11.Region] or list[object]
    :param rates: The neutral mutation rate, selected mutation rate, and
                  total recombination rate, respectively.
                  See below for more details.
    :type rates: list or fwdpy11.MutationAndRecombinationRates
    :param demography: The demographic model to simulate
    :type demography: object
    :param simlen: The number of time steps to evolve
    :type simlen: int
    :param prune_selected: If ``True``, remove selected fixations from
                           the population when they are first detected
                           as fixed.
    :type prune_selected: bool
    :param allow_residual_selfing: If ``True``, then an individual may
                                   be chosen twice as a parent.
                                   If ``False``, two distinct parents
                                   are required.
    :type allow_residual_selfing: bool

    .. warning::

        When ``allow_residual_selfing`` is ``False``, an exception
        will be raised if the parental deme has a size of 1.
        The reason is that outcrossing is not possible in order
        to generate the offspring.

    .. note::

        To initialize the ``rates`` field, we require an instance of
        fwdpy11.MutationAndRecombinationRates or a list of length three (3)
        specifying the three rates.
        The two mutation rates must be non-negative floats. For
        the recombination rate, the third value must also
        be a non-negative float if all objects in ``recrates``
        are instances of :class:`fwdpy11.Region`. However,
        if they are instead instances of any mixture of
        :class:`fwdpy11.PoissonInterval`, :class:`fwdpy11.PoissonPoint`,
        :class:`fwdpy11.BinomialPoint`, or :class:`fwdpy11.BinomialInterval`,
        then the final element in ``rates`` must be ``None``.
        See the :ref:`section <geneticmaps_vignette>` on setting
        recombination rates for details.


    .. versionadded:: 0.1.1

    .. versionchanged:: 0.2.0
        Changed this from a horrible class hierarchy
        into a much simpler, single class.

    .. versionchanged:: 0.6.0

        Updated to support `fwdpy11.DiscreteDemography`

    .. versionchanged:: 0.8.0

        Refactored class internals using ``attrs``.
        Mutation and recombination rates now
        stored in :class:`fwdpy11.MutationAndRecombinationRates`

    .. versionchanged:: 0.14.0

        Remove deprecated kwargs pself and popsizes.

    .. versionchanged:: 0.20.0

        Add `allow_residual_selfing`
    """

    nregions = attr.ib(factory=list)
    sregions = attr.ib(factory=list)
    recregions = attr.ib(factory=list)
    rates: MutationAndRecombinationRates = attr.ib(converter=_convert_rates)
    gvalue = attr.ib(default=None)
    demography: typing.Union[ForwardDemesGraph,
                             DemographicModelDetails] = attr.ib(default=None)
    simlen: int = attr.ib(converter=int, default=0)
    prune_selected: bool = attr.ib(default=True)
    allow_residual_selfing: bool = attr.ib(default=True)

    @nregions.validator
    def validate_nregions(self, attribute, value):
        for i in value:
            attr.validators.instance_of(fwdpy11.Region)(self, attribute, i)

    @sregions.validator
    def validate_sregions(self, attribute, value):
        for i in value:
            attr.validators.instance_of(fwdpy11.Sregion)(self, attribute, i)
            try:
                if i.shape != self.gvalue.shape:
                    e = "Sregion and genetic value "
                    "dimension mismatch: {} {}, {} {}".format(
                        type(i), i.shape, type(self.gvalue), self.gvalue.shape
                    )
                    raise ValueError(e)
            except AttributeError:
                for g in self.gvalue:
                    if i.shape != g.shape:
                        e = "Sregion and genetic value "
                        "dimension mismatch: {} {}, {} {}".format(
                            type(i), i.shape, type(self.gvalue), g.shape
                        )
                        raise ValueError(e)

    @recregions.validator
    def validate_recregions(self, attribute, value):
        if len(value) > 0 and all([isinstance(i, fwdpy11.Region) for
                                   i in value]):
            warnings.warn(
                "using a list of Regions for recregions is deprecated",
                UserWarning, stacklevel=2)
        try:
            for i in value:
                attr.validators.instance_of(fwdpy11.Region)(self, attribute, i)
        except TypeError:
            try:
                for i in value:
                    valid = False
                    for k in [fwdpy11.PoissonCrossoverGenerator,
                              fwdpy11.NonPoissonCrossoverGenerator]:
                        if isinstance(i, k):
                            valid = True
                            break

                    if not valid:
                        raise TypeError(f"invalid recregion type {type(i)}")
                if not all([i.discrete for i in value]) and not all(
                    [not i.discrete for i in value]
                ):
                    warnings.warn(
                        "genetic map has a mix of discrete=True"
                        "and discrete=False", stacklevel=2
                    )

            except TypeError:
                raise

    @rates.validator
    def rates_validator(self, attribute, value):
        if value.recombination_rate is None:
            for i in self.recregions:
                if not isinstance(i, fwdpy11.PoissonCrossoverGenerator) and \
                        not isinstance(i,
                                       fwdpy11.NonPoissonCrossoverGenerator):
                    raise ValueError(
                        f"recombination rate of {value.recombination_rate}"
                        " must be paired with"
                        " instances of fwdpy11.GeneticMapUnit"
                    )
        else:
            for i in self.recregions:
                if not isinstance(i, fwdpy11.Region):
                    raise ValueError(
                        f"recombination rate of {value.recombination_rate}"
                        " must be paired with"
                        " instances of fwdpy11.Region"
                    )

    @gvalue.validator
    def validate_gvalue(self, attribute, value):
        try:
            for i in value:
                attr.validators.instance_of(fwdpy11.DiploidGeneticValue)(
                    self, attribute, i
                )
        except TypeError:
            try:
                attr.validators.instance_of(fwdpy11.DiploidGeneticValue)(
                    self, attribute, value
                )
            except TypeError:
                raise

    @demography.validator
    def validate_demography(self, attribute, value):
        if value is None:
            warnings.warn("No demographic model specified."
                          " This will be a hard error in a future release.",
                          DeprecationWarning)
            return

        if isinstance(value, fwdpy11.ForwardDemesGraph):
            return

        if isinstance(value, DemographicModelDetails):
            return
        raise ValueError(f"Unknown type for {attribute}: {type(value)}")

    @simlen.validator
    def validate_simlen(self, attribute, value):
        if value <= 0:
            raise ValueError(
                f"{attribute} must be >= 0, but we got {value} instead")



#
# Copyright (C) 2019 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import typing

import demes
import demes.demes


from fwdpy11._types.demographic_model_details import DemographicModelDetails


def from_demes(
    dg: typing.Union[str, demes.Graph], burnin: int = 10, *,
    round_non_integer_sizes: typing.Optional[bool] = None
) -> DemographicModelDetails:
    """
    Build a :class:`fwdpy11.DemographicModelDetails` object using demes.
    The deme graph can either be a demes Graph object or a string as the
    filepath to a demes-specifiend YAML demography.

    :param dg: The demes Graph to convert.
    :type dg: demes.Graph or str
    :param burnin: A factor for how many generations to burn in the simulation.
        For a typical demography with a single root, the number of generations
        of burn in is `burnin` times the root deme's population size. For
        models with multiple root demes joined by migration, that population
        size is determined as the size of the metapopulation.
    :type burnin: int

    :rtype: :class:`fwdpy11.DemographicModelDetails`

    .. versionadded:: 0.14.0
    """
    from ._functions import demography_from_demes

    demog = demography_from_demes(
        dg, burnin, round_non_integer_sizes=round_non_integer_sizes)
    return demog



# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your meanion) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.

# Main goal is to test pickling

import unittest

import numpy as np

import fwdpy11


class testGeneticValueIsFitness(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.g = fwdpy11.GeneticValueIsFitness(1)

    def testPickle(self):
        import pickle

        p = pickle.dumps(self.g, -1)
        up = pickle.loads(p)
        self.assertEqual(type(up), type(self.g))
        self.assertEqual(self.g.shape, up.shape)


class testGSS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.VS = 1.0
        self.opt = 0.0
        self.g = fwdpy11.GaussianStabilizingSelection.single_trait(
            [fwdpy11.Optimum(VS=self.VS, optimum=self.opt, when=None)])

    def testPickle(self):
        import pickle

        p = pickle.dumps(self.g, -1)
        up = pickle.loads(p)
        self.assertEqual(up.optima[0].VS, self.VS)
        self.assertEqual(up.optima[0].optimum, self.opt)

    def test_mapping(self):
        self.assertEqual(self.g.maps_to_fitness, False)
        self.assertEqual(self.g.maps_to_trait_value, True)


class testGSSmo(unittest.TestCase):
    @classmethod
    def setUp(self):
        Opt = fwdpy11.Optimum
        self.optima = [
            Opt(when=0, optimum=0.0, VS=1.0),
            Opt(when=100, optimum=1.0, VS=1.0),
        ]
        self.g = fwdpy11.GaussianStabilizingSelection.single_trait(self.optima)

    def testPickle(self):
        import pickle

        p = pickle.dumps(self.g)
        up = pickle.loads(p)
        # self.assertEqual(up.VS, 1.0)
        # self.assertEqual(up.optimum, 0.0)
        for i, j in zip(up.optima, self.optima):
            self.assertEqual(i.when, j.when)
            self.assertEqual(i.optimum, j.optimum)
            self.assertEqual(i.VS, j.VS)

    def test_mapping(self):
        self.assertEqual(self.g.maps_to_fitness, False)
        self.assertEqual(self.g.maps_to_trait_value, True)


class testMultivariateGSSmo(unittest.TestCase):
    @classmethod
    def setUp(self):
        optima = np.array([0.0, 0.0, 1.0, 1.0]).reshape(2, 2)
        timepoints = [0, 100]
        po = []
        for i, t in enumerate(timepoints):
            po.append(fwdpy11.PleiotropicOptima(
                when=t, optima=optima[i, :], VS=1.0))
        self.mvgssmo = fwdpy11.GaussianStabilizingSelection.pleiotropy(po)

    def test_pickle(self):
        import pickle

        p = pickle.dumps(self.mvgssmo)
        up = pickle.loads(p)
        self.assertEqual(self.mvgssmo, up)

    def test_mapping(self):
        self.assertEqual(self.mvgssmo.maps_to_fitness, False)
        self.assertEqual(self.mvgssmo.maps_to_trait_value, True)


def test_gaussian_stabilizing_selection_single_trait():
    gss = fwdpy11.GaussianStabilizingSelection.single_trait(
        [fwdpy11.Optimum(0.0, 10.0, None)])
    d = gss.asdict()
    gss2 = gss.fromdict(d)
    assert gss == gss2
    s = str(d)
    gss2 = gss.fromdict(eval(s))
    assert gss == gss2


def test_gaussian_stabilizing_selection_pleiotropy():
    gss = fwdpy11.GaussianStabilizingSelection.pleiotropy(
        [fwdpy11.PleiotropicOptima([0, 0, 0, 0], 1.0, 0)])
    d = gss.asdict()
    gss2 = gss.fromdict(d)
    assert gss == gss2
    s = str(d)
    gss2 = gss.fromdict(eval(s))
    assert gss == gss2


if __name__ == "__main__":
    unittest.main()



import unittest

import fwdpy11
import fwdpy11.custom_genetic_value_decorators


@fwdpy11.custom_genetic_value_decorators.genetic_value_noise_default_clone
class InheritedNoise(fwdpy11.GeneticValueNoise):
    def __init__(self):
        self.generation = 0
        fwdpy11.GeneticValueNoise.__init__(self)

    def __call__(self, data) -> float:
        return data.parent1_metadata.e + self.generation

    def update(self, pop):
        self.generation = pop.generation


class TestInheritNoise(unittest.TestCase):
    """
    The class InheritedNoise passes the
    sum of all generations into the "e" field of
    offspring metadata.  Thus, if offspring metadata
    can see parental metadata, then we can test that the
    final alive individuals (in a sim of non-overlapping
    generations) all have e fields that are the sum of
    the generations that have occurred.
    """

    @classmethod
    def setUp(self):
        self.rng = fwdpy11.GSLrng(42)
        self.pop = fwdpy11.DiploidPopulation([100, 100, 100], 1.0)
        self.pdict = {
            "nregions": [],
            "sregions": [],
            "recregions": [],
            "rates": (0, 0, 0),
            "demography": fwdpy11.ForwardDemesGraph.tubes(
                [100] * 3, 1),
            "simlen": 3,
            "gvalue": fwdpy11.Additive(
                2.0, fwdpy11.GaussianStabilizingSelection.single_trait(
                    [fwdpy11.Optimum(optimum=0.0, VS=1.0, when=0)]), InheritedNoise()
            ),
        }

    def test_noise_values_tree_sequences(self):
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        self.assertEqual(self.pop.generation, 3)
        self.assertTrue(
            all([i.e == 6.0 for i in self.pop.diploid_metadata]) is True)


if __name__ == "__main__":
    unittest.main()



"""
tskit flags are unsigned 32 bit integers.

The first 16 bits are reserved for internal
use by tskit, so 16 and on are available
to us.

See here for details:
    https://tskit.readthedocs.io/en/latest/data-model.html?highlight=reserved#table-definitions

"""
INDIVIDUAL_IS_ALIVE = 2 ** 16
INDIVIDUAL_IS_PRESERVED = 2 ** 17
INDIVIDUAL_IS_FIRST_GENERATION = 2 ** 18



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

import fwdpy11


class TestFwdppExceptions(unittest.TestCase):
    def test_tables_error(self):
        with self.assertRaises(fwdpy11.TablesError):
            raise fwdpy11.TablesError("this is a tables error")

    def test_tables_error_from_cpp(self):
        import fwdpy11._fwdpy11
        with self.assertRaises(fwdpy11.TablesError):
            fwdpy11._fwdpy11._throw_TablesError()

    def test_samples_error(self):
        with self.assertRaises(fwdpy11.SamplesError):
            raise fwdpy11.SamplesError("this is a tables error")

    def test_samples_error_from_cpp(self):
        import fwdpy11._fwdpy11
        with self.assertRaises(fwdpy11.SamplesError):
            fwdpy11._fwdpy11._throw_SamplesError()


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import copy
import typing

import fwdpy11
import tskit  # type: ignore

TopLevelMetadata = tskit.metadata.MetadataSchema(
    {
        "codec": "json",
        "title": "Top-level metadata for table collection/tree sequence.",
        "type": "object",
        "properties": {
            "generation": {
                "type": "integer",
                "description": "The value of pop.generation at the time data"
                "were exported to tskit",
            },
            "model_params": {
                "type": ["string", "object"],
                "description": "One or more fwdpy11.ModelParams instances.",
            },
            "seed": {
                "type": "integer",
                "description": "Random number seed."
                "This is optional because a random number generator"
                "may be called prior to simulation, thus making the"
                "initial seed not capable of reproducing the simulation",
            },
            "demes_graph": {
                "type": "object",
                "description": "A demographic model specified using demes."
                "This information will be redundant with that stored in model_params,"
                "but it may be useful as it allows reconstruction of the YAML file"
                "from the tree sequence.",
            },
            "data": {
                "type": ["object", "string"],
                "description": "This field is reserved for the user to fill.",
            },
        },
        "required": ["generation"],
    }
)

IndividualDiploidMetadata = tskit.metadata.MetadataSchema(
    {
        "codec": "struct",
        "type": "object",
        "properties": {
            "g": {"type": "number", "binaryFormat": "d"},
            "e": {"type": "number", "binaryFormat": "d"},
            "w": {"type": "number", "binaryFormat": "d"},
            "sex": {"type": "number", "binaryFormat": "i"},
            "deme": {"type": "number", "binaryFormat": "i"},
            "label": {"type": "number", "binaryFormat": "Q"},
            "parents": {
                "type": "array",
                "items": {"type": "number", "binaryFormat": "Q"},
                "arrayLengthFormat": "H",
            },
            "geography": {
                "type": "array",
                "items": {"type": "number", "binaryFormat": "d"},
                "arrayLengthFormat": "H",
            },
            "nodes": {
                "type": "array",
                "items": {"type": "number", "binaryFormat": "i"},
                "arrayLengthFormat": "H",
            },
        },
    }
)

PopulationMetadata = tskit.metadata.MetadataSchema(
    {
        "codec": "json",
        "type": "object",
        "name": "Population metadata",
        "properties": {"name": {"type": "string"}},
    }
)

# NOTE: we could use JSON schema here and
# have the array fields not included in "required".
# However:
# 1. That is much slower.
# 2. This metadata contains floating point values.
#    We don't want to lose precision in case anyone
#    wants to try to recalculate diploid phenotypes/fitnesses

MutationMetadata = tskit.metadata.MetadataSchema(
    {
        "codec": "struct",
        "type": ["object", "null"],
        "name": "Mutation metadata",
        "properties": {
            "s": {"type": "number", "binaryFormat": "d"},
            "h": {"type": "number", "binaryFormat": "d"},
            "origin": {"type": "number", "binaryFormat": "i"},
            "neutral": {"type": "number", "binaryFormat": "?"},
            "label": {"type": "number", "binaryFormat": "H"},
            "key": {"type": "number", "binaryFormat": "Q"},
        },
        "additionalProperties": False,
    }
)

_MutationMetaWithVectorsDict = copy.deepcopy(MutationMetadata.schema)

_MutationMetaWithVectorsDict["name"] = "Mutation metadata with vectors"
_MutationMetaWithVectorsDict["properties"]["esizes"] = {
    "type": "array",
    "items": {"type": "number", "binaryFormat": "d"},
}
_MutationMetaWithVectorsDict["properties"]["heffects"] = {
    "type": "array",
    "items": {"type": "number", "binaryFormat": "d"},
}

MutationMetadataWithVectors = tskit.metadata.MetadataSchema(
    _MutationMetaWithVectorsDict
)


def generate_individual_metadata(
    metadata: fwdpy11._fwdpy11.DiploidMetadata,
) -> typing.Dict:
    d = {
        "g": metadata.g,
        "e": metadata.e,
        "w": metadata.w,
        "geography": [i for i in metadata.geography],
        "parents": [i for i in metadata.parents],
        "nodes": [i for i in metadata.nodes],
        "sex": metadata.sex,
        "deme": metadata.deme,
        "label": metadata.label,
    }
    return d


def determine_mutation_metadata_schema(
    mutations: fwdpy11._fwdpy11.MutationVector,
) -> tskit.metadata.MetadataSchema:
    if len(mutations) == 0 or len(mutations[0].esizes) == 0:
        return MutationMetadata

    return MutationMetadataWithVectors


def generate_mutation_metadata(
    mr: fwdpy11._fwdpy11.MutationRecord, mutations: fwdpy11._fwdpy11.MutationVector
) -> typing.Dict:
    m = mutations[mr.key]
    d = {
        "s": m.s,
        "h": m.h,
        "origin": m.g,
        "label": m.label,
        "neutral": int(m.neutral),
        "key": mr.key,
    }
    if len(m.esizes) > 0:
        d["esizes"] = list(m.esizes)
        d["heffects"] = list(m.heffects)
    return d



# Convert a yaml Demes model to a fully qualified json model and write
# to stdout.
import sys
import json
from ruamel.yaml import YAML

import demes_parser as parser


if __name__ == "__main__":

    yaml = YAML(typ="safe")
    with open(sys.argv[1], encoding="utf-8") as source:
        data = yaml.load(source)
    graph = parser.parse(data)
    print(json.dumps(graph.as_json_dict(), indent=2))



from typing import IO, Dict, Iterable, List, Optional, Tuple, Union

import demes
import fwdpy11._types
import fwdpy11.tskit_tools._dump_tables_to_tskit
import numpy as np
import tskit  # type: ignore

from .._fwdpy11 import DiploidGenotype, DiploidMetadata, ll_DiploidPopulation
from .model_params import ModelParams
from .new_mutation_data import NewMutationData
from .population_mixin import PopulationMixin
from .table_collection import TableCollection


class DiploidPopulation(ll_DiploidPopulation, PopulationMixin):
    """
    A population of diploid individuals.

    To initialize a population of size N and genome length L:

    fwdpy11.DiploidPopulation(N, L)

    To initialize a population with two demes of initial
    sizes N1 and N2 and genome length L:

    fwdpy11.DiploidPopulation([N1, N2], L)

    .. versionchanged:: 0.16.0

        Added __copy__ and __deepcopy__.
        In general, the correct way to copy
        instances if this class is via :func:`copy.deepcopy`.
        Calling :func:`copy.copy` will copy some of the underlying
        C++ objects, but the two objects will share the same
        table collection.  This situation of sharing is almost certainly
        not what one wants.
    """

    def __init__(
        self,
        N: Union[int, List[int]],
        length: float,
        *,
        ll_pop: Optional[ll_DiploidPopulation] = None,
    ):
        if ll_pop is None:
            super(DiploidPopulation, self).__init__(N, length)
        else:
            super(DiploidPopulation, self).__init__(ll_pop)

        self._pytables = TableCollection(self._tables)

    def __copy__(self):
        ll_pop = super(DiploidPopulation, self).__copy__()
        return self.__class__(0, 0.0, ll_pop=ll_pop)

    def __deepcopy__(self, memo):
        ll_pop = super(DiploidPopulation, self).__deepcopy__(memo)
        return self.__class__(0, 0.0, ll_pop=ll_pop)

    @classmethod
    def create_from_tskit(cls, ts: tskit.TreeSequence,
                          *, import_mutations=False,
                          import_individuals=False):
        """
        Create a new object from an tskit.TreeSequence

        :param ts: A tree sequence from tskit
        :type ts: tskit.TreeSequence
        :param import_mutations: if `True`, import mutations and sites
                                 from the tree sequence
        :type import_mutations: bool

        :return: A population object with an initialized
                 :class:`fwdpy11.TableCollection`
        :rtype: :class:`fwdpy11.DiploidPopulation`

        .. note::

            In general, initializing a population using
            the output from a coalescent simulation is
            a tricky business.  There are issues of
            parameter scaling and the appropriateness
            of the coalescent model itself. A key issue
            is that your input tree sequence must have
            node times in the correct time units! (Generations,
            for example.)

        .. warning::

            Support for importing mutations is currently limited to
            two cases:

            1. The tree seq has mutations entirely from tskit's API
               (see :ref:`import_mutations_from_tskit_vignette`).
            2. The tree seq has mutations entirely from a previous run
               of fwdpy11.


        .. versionadded:: 0.2.0

        .. versionchanged:: 0.19.8

           Correctly handle two cases:

           1. The tree seq has mutations entirely from tskit's API
              (see :ref:`import_mutations_from_tskit_vignette`).
           2. The tree seq has mutations entirely from a previous run
              of fwdpy11.

        """
        generation = 0
        if len(ts.metadata) > 0:
            if "generation" in ts.metadata:
                generation = ts.metadata["generation"]
        ll = ll_DiploidPopulation._create_from_tskit(ts, generation)
        if import_mutations is True:
            import fwdpy11.tskit_tools
            for mutation in ts.mutations():
                # if mutation.metadata["origin"] < 0.0:
                #     msg = "all origin fields in mutation metadata"
                #     msg += "must be non-negative"
                #     raise ValueError(msg)
                if not mutation.time.is_integer():
                    raise ValueError("mutation times cannot contain decimals")
            mutation_nodes = [i.node for i in ts.mutations()]
            decoded_md = fwdpy11.tskit_tools.decode_mutation_metadata(ts)
            try:
                # The case of the tree seq has all mutations added
                # to it via tskit API
                for i, j in zip(ts.mutations(), decoded_md):
                    if j is not None:
                        if i.time != j.g:
                            msg = f"metadata origin time {j.g} != "
                            msg += f"mutation time {i.time} in tables"
                            raise ValueError(msg)
            except ValueError:
                # The case of the tree seq has all of its mutations
                # from a pervious fwdpy11 run
                keys = set([i.key for i in decoded_md])
                for i in ts.mutations():
                    md = i.metadata
                    key = (ts.site(i.site).position, md['s'], md['origin'])
                    if key not in keys:
                        raise ValueError(f"{key}")

            mvec = fwdpy11._fwdpy11.MutationVector()
            for i in decoded_md:
                if i is not None:
                    mvec.append(i)
            times = [int(i.time) - ll._generation for i in ts.mutations()]
            ll._set_mutations(mvec, mutation_nodes, times)
        if import_individuals is True:
            import fwdpy11.tskit_tools

            def liftover(input, output):
                for i in range(len(input)):
                    output['g'][i] = input[i].g
                    output['w'][i] = input[i].w
                    output['e'][i] = input[i].e
                    output['sex'][i] = input[i].sex
                    output['deme'][i] = input[i].deme
                    output['parents'][i] = input[i].parents
                    output['nodes'][i] = input[i].nodes
                    output['geography'][i] = input[i].geography
                    output['label'][i] = input[i].label
            decoded_md = fwdpy11.tskit_tools.decode_individual_metadata(ts)
            # Take a writeable VIEW of the C++-side data...
            md = np.array(ll._diploid_metadata, copy=False)
            alive_md = [i for i in decoded_md if i.alive is True]
            ancient_sample_md = [i for i in decoded_md if i.alive is False]
            ll._ancient_sample_metadata._resize(len(ancient_sample_md))
            amd = np.array(ll._ancient_sample_metadata, copy=False)
            # ...overwrite that data here
            liftover(alive_md, md)
            liftover(ancient_sample_md, amd)
        return cls(0, 0.0, ll_pop=ll)

    @ classmethod
    def load_from_file(cls, filename: str):
        """
        Load population from the output of
        :func:`fwdpy11.DiploidPopulation.dump_to_file`.

        :param filename: A file name
        :type filename: str
        """
        ll = ll_DiploidPopulation._load_from_file(filename)
        return cls(0, 0.0, ll_pop=ll)

    @ classmethod
    def load_from_pickle_file(cls, filename: IO):
        """
        Read in a pickled population from a file.
        The file muse have been generated by
        a call to :func:`fwdpy11.DiploidPopulation.pickle_to_file`.

        :param f: A handle to a file opened in 'rb' mode.

        .. versionadded: 0.3.0

        """
        ll = ll_DiploidPopulation._load_from_pickle_file(filename)
        return cls(0, 0.0, ll_pop=ll)

    @ property
    def alive_nodes(self) -> np.ndarray:
        """
        List of alive nodes corresponding to individuals.

        .. versionadded:: 0.5.3
        """
        md = np.array(self.diploid_metadata, copy=False)
        return md["nodes"].flatten()

    @ property
    def ancient_sample_nodes(self) -> np.ndarray:
        """
        Return an array of nodes associated with preserved/ancient samples.

        Alias for :attr:`fwdpy11.DiploidPopulation.preserved_nodes`.

        .. versionadded:: 0.13.0
        """
        return self.preserved_nodes

    @ property
    def ancient_sample_metadata(self) -> Iterable[DiploidMetadata]:
        """Supports buffer protocol"""
        return self._ancient_sample_metadata

    @ property
    def diploids(self) -> Iterable[DiploidGenotype]:
        """Supports buffer protocol"""
        return self._diploids

    @ property
    def diploid_metadata(self) -> Iterable[DiploidMetadata]:
        """Supports buffer protocol"""
        return self._diploid_metadata

    @ property
    def preserved_nodes(self) -> np.ndarray:
        """
        Return an array of nodes associated with preserved/ancient samples.

        :rtype: numpy.ndarray

        .. versionadded:: 0.13.0

        """
        return np.array(self.ancient_sample_metadata,
                        copy=False)["nodes"].flatten()

    @ property
    def tables(self) -> TableCollection:
        """Access the :class:`fwdpy11.TableCollection`"""
        return self._pytables

    def _get_times(self):
        amd = np.array(self.ancient_sample_metadata, copy=False)
        nodes = np.array(self.tables.nodes, copy=False)
        times = nodes["time"][amd["nodes"][:, 0]]
        utimes = np.unique(times)
        return times, utimes

    def deme_sizes(self, as_dict=False):
        """
        Return the number of individuals in each deme.

        :param as_dict: If True, return results as a dict.
                        If False, numpy arrays are returned

        The default behavior of this function is equivalent to:

        .. code-block:: python

            md = numpy.array(self.diploid_metadata, copy=False)
            return numpy.unique(md['deme'], return_counts=True)

        .. versionadded:: 0.6.0
        """
        md = np.array(self.diploid_metadata, copy=False)
        deme_sizes = np.unique(md["deme"], return_counts=True)
        if as_dict is False:
            return deme_sizes
        return {i: j for i, j in zip(deme_sizes[0], deme_sizes[1])}

    def dump_tables_to_tskit(
        self,
        *,
        model_params: Optional[Union[ModelParams,
                                     Dict[str, ModelParams]]] = None,
        demes_graph: Optional[demes.Graph] = None,
        population_metadata: Optional[Dict[int, object]] = None,
        data: Optional[object] = None,
        seed: Optional[int] = None,
        parameters: Optional[Dict] = None,
        destructive=False,
    ) -> tskit.TreeSequence:
        """
        Dump the population's TableCollection into
        an tskit TreeSequence

        :param model_params: Model parameters to be stored as top-level
                             metadata
        :type model_params: :class:`fwdpy11.ModelParams` or :class:`dict`

        :param demes_graph: A demographic model specified via `demes`.
        :type demes_graph: :class:`demes.Graph`

        :param population_metadata: A mapping from integer id of a
                                    deme/population to metadata
        :type population_metadata: dict

        :param data: User-generated data to add to top-level metadata
        :type data: object

        :param seed: Random number seed for top-level metadata.
                     If not `None`, `seed` must be `>= 0`.
        :type seed: int

        :param parameters: The simulation parameters for the provenance table.
        :type parameters: None or dict

        :param destructive: If `True`, delete data held by the current
                            instance.
        :type destructive: bool

        :rtype: tskit.TreeSequence

        For examples, see :ref:`tskitconvert_vignette`.

        .. warning::

            If `destructive` is `True`, further opertations on the
            current instance should be considered undefined behavior
            that may lead to a crash.

        .. versionchanged:: 0.8.2

            Added `parameters`.
            Generate provenance information for return value.
            The provenance information is validated using
            :func:`tskit.validate_provenance`, which may
            raise an exception.

        .. versionchanged:: 0.10.0

            Use tskit metadata schema.
            Mutation time is now stored in the tskit.MutationTable column.
            Origin time of mutations is part of the metadata.

        .. versionchanged:: 0.14.0

            Added `destructive` option.

        .. versionchanged:: 0.15.0

            Added `model_params`, `demes_graph`, `population_metadata`,
            `data` keyword args.

        .. versionchanged:: 0.15.2

            Fixed bug that could generate a :class:`tskit.PopulationTable`
            with an incorrect number of rows.

        .. versionchanged:: 0.17.0

            The `wrapped` keyword argument is deprecated.

        .. versionadded:: 0.19.0

            Removed deprecated `wrapped` keyword argument.

        """
        return fwdpy11.tskit_tools._dump_tables_to_tskit._dump_tables_to_tskit(
            self,
            model_params=model_params,
            demes_graph=demes_graph,
            population_metadata=population_metadata,
            data=data,
            seed=seed,
            parameters=parameters,
            destructive=destructive,
        )

    def dump_to_file(self, filename: str):
        """
        Write a population to a file in binary format.
        """
        self._dump_to_file(filename)

    def pickle_to_file(self, filename: IO):
        """
        Pickle the population to an open file.

        This function may be preferred over
        the direct pickling method because it uses less
        memory.  It is, however, slower.

        To read the population back in, you must call
        :func:`fwdpy11.DiploidPopulation.load_from_pickle_file`.

        :param f: A handle to an open file

        .. versionadded:: 0.3.0

        """
        self._pickle_to_file(filename)

    def sample_timepoints(
        self, include_alive=True
    ):
        """
        Return an iterator over all sample time points.
        The iterator yields time, nodes, and metadata.

        :param include_alive: If True, include currently-alive individuals.

        .. versionadded :: 0.5.1

        .. versionchanged:: 0.5.3

           Monkey-patched into pybind11 class
        """
        amd = np.array(self.ancient_sample_metadata, copy=False)
        amd.flags.writeable = False
        times, utimes = self._get_times()
        for uti in utimes:
            mdidx = np.where(times == uti)[0]
            sample_nodes = amd["nodes"][mdidx].flatten()
            sample_nodes.flags.writeable = False
            mdslice = amd[mdidx][:]
            mdslice.flags.writeable = False
            yield uti, sample_nodes, mdslice

        if include_alive is True:
            md = np.array(self.diploid_metadata, copy=False)
            nodes = md["nodes"].flatten()
            md.flags.writeable = False
            nodes.flags.writeable = False
            yield self.generation, nodes, md

    def add_mutation(
        self,
        rng: fwdpy11.GSLrng,
        *,
        window: Tuple[float, float] = None,
        ndescendants: int = 1,
        deme: Optional[int] = None,
        data: NewMutationData = None,
    ) -> Optional[int]:
        """
        Add a new mutation to the population.

        .. versionadded:: 0.16.0

        :param rng: Random number generator
        :type rng: fwdpy11.GSLrng

        The following arguments are keyword-only:

        :param window: A window [left, right) within which to place
                       the mutation. The default is `None`, meaning
                       the window is the entire genome.
        :type window: tuple[float, float]
        :param ndescendants: The number of alive nodes carrying the new
                             mutation. Default is `1`, implying that a
                             singleton mutation will be generated.
        :type ndescendants: int
        :param deme: The deme in which to place the new mutation
                     The default is `None`, meaning that alive node demes are
                     not considered.
        :type deme: int
        :type data: The details of the new mutation
        :type data: fwdpy11.NewMutationData

        :returns: The key of the new mutation.
                  (The index of the variant in
                  :attr:`DiploidPopulation.mutations`.)

        Implementation details:

        * A set of nodes are found within `window` that are ancestral to
          exactly `ndescendants` alive nodes.
        * From this list of candidate nodes, one is chosen randomly
          to be the node where we place the new mutation.
        * If the node's parent's time is NULL, then the mutations' origin
          time is the first 64 bit integer ancestral to the node time,
          which could be the node time itself.
          Otherwise, a 64 bit integer is chosen uniformly between the
          node time and the node's parent's time.
        * If `deme` is None, any set of `ndescendants` will be considered.
        * If `deme` is `>= 0`, all `ndescendants` alive nodes must be
          from that deme.
        * If the `deme`/`ndescendants` requirements cannot be satisified,
          the function returns `None`.
        * The genetic values of individuals are not updated by this function.
          Such updates will happen when the population begins evolution
          forwards in time.

        Exceptions:

        :raises ValueError: for bad input.
        :raises RuntimeError: if this function is applied during a simulation
        :raises RuntimeError: if internal checks fail.
        """
        if window is None:
            _window = (0.0, self.tables.genome_length)
        else:
            _window = window

        if deme is None:
            _deme = -1
        else:
            _deme = deme

        if ndescendants is None:
            raise ValueError(f"ndescendants must be > 0: got {ndescendants}")

        from fwdpy11._fwdpy11 import _add_mutation

        key = _add_mutation(
            rng, _window[0], _window[1], ndescendants, _deme, data, self
        )
        if key == np.iinfo(np.uint64).max:
            return None
        return key



#
# Copyright (C) 2017-2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

from typing import List, Tuple

import numpy as np

from .._fwdpy11 import ll_DataMatrix, StateMatrix


class DataMatrix(ll_DataMatrix):
    """
    Represent a sample from a population in a matrix format.

    There are two possible representations of the data:

    1. As a genotype matrix, where individuals are encoded a 0,1, or 2
    copies of the derived mutation. There is one column per diploid here,
    and one row per variable site.

    2. As a haplotype matrix, with two columns per diploid, and each
    column containing a 0 (ancestral) or 1 (derived) label. Each row
    represents a variable site.

    .. versionchanged:: 0.2.0

        Changed layout to row = variable site.
        Changed to match fwdpp 0.7.0 layout where the neutral
        and selected data are represented as a
        :class:`fwdpy11.StateMatrix`
    """

    def __init__(self, *args):
        super(DataMatrix, self).__init__(*args)

    @property
    def ncol(self) -> int:
        """Sample size of the matrix"""
        return self._ncol

    @property
    def neutral(self) -> StateMatrix:
        """:class:`fwdpy11.StateMatrix` for neutral variants"""
        return self._neutral

    @property
    def neutral_keys(self) -> List[int]:
        """Keys for neutral mutations used to generate matrix."""
        return self._neutral_keys

    @property
    def selected(self) -> StateMatrix:
        """:class:`fwdpy11.StateMatrix` for selected variants"""
        return self._selected

    @property
    def selected_keys(self) -> List[int]:
        """Keys for selected mutations used to generate matrix."""
        return self._selected_keys

    def merge(self) -> Tuple[np.ndarray, List[int]]:
        """
        Merge the neutral and selected data.

        :rtype: tuple
        :returns: a :class:`numpy.ndarray` composed of
                  neutral and selected mutations.  Rows (mutations)
                  are sorted by position.  The second tuple element
                  contains the mutation keys.

        .. versionadded:: 0.6.1
        """
        nneutral = len(self.neutral_keys)
        nselected = len(self.selected_keys)

        if nneutral == 0 and nselected > 0:
            return np.array(self.selected), self.selected_keys
        elif nneutral > 0 and nselected == 0:
            return np.array(self.neutral), self.neutral_keys

        positions = np.concatenate((self.neutral.positions, self.selected.positions))
        sorted_pos_indexes = np.argsort(positions)
        merged = np.concatenate(
            (np.array(self.neutral), np.array(self.selected)), axis=0
        )
        keys = np.concatenate((self.neutral_keys, self.selected_keys))
        return merged[sorted_pos_indexes, :], keys[sorted_pos_indexes]

    def _return_sorted_matrix(self, sm, k):
        pos = sm.positions
        posi = np.argsort(pos)
        return np.array(sm, copy=False)[posi, :], k

    def neutral_matrix(self) -> Tuple[np.ndarray, List[int]]:
        """
        :returns: Neutral mutations and keys
        :rtype: tuple

        The neutral matrix is a :class:`numpy.ndarray`

        .. versionadded:: 0.6.1
        """
        return self._return_sorted_matrix(self.neutral, self.neutral_keys)

    def selected_matrix(self) -> Tuple[np.ndarray, List[int]]:
        """
        :returns: Selected mutations and keys
        :rtype: tuple

        The selected matrix is a :class:`numpy.ndarray`

        .. versionadded:: 0.6.1
        """
        return self._return_sorted_matrix(self.selected, self.selected_keys)



#
# Copyright (C) 2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import copy
import typing

import attr
import fwdpy11
import numpy as np
from fwdpy11.class_decorators import attr_class_to_from_dict
from fwdpy11.conditional_models import (
    AddMutationFailure,
    AlleleCount,
    AlleleCountRange,
    AncientSamplePolicy,
    ConditionalModelOutput,
    EvolveOptions,
    FrequencyRange,
    NewMutationParameters,
    OutOfAttempts,
    SimulationStatus,
    _non_negative_value,
)


@attr_class_to_from_dict
@attr.s(auto_attribs=True)
class _InternalSweepEvolveOptions:
    track_mutation_counts: bool = True


@attr.s(auto_attribs=True)
class _ProgressMonitor:
    index: int = attr.ib(
        validator=[
            attr.validators.instance_of(int),
            _non_negative_value,
        ]
    )
    key: typing.Tuple[float, float, int]
    return_when_stopping_condition_met: bool
    status: SimulationStatus

    def __call__(self, pop: fwdpy11.DiploidPopulation, _) -> bool:
        if (
            self.status == SimulationStatus.Success
            and self.return_when_stopping_condition_met is True
        ):
            return True
        return self.status == SimulationStatus.Restart


@attr.s(auto_attribs=True)
class _MutationPresent:
    when: int = attr.ib(
        validator=[attr.validators.instance_of(int), _non_negative_value]
    )
    until: typing.Optional[int] = attr.ib(attr.validators.optional(int))  # type: ignore

    def __attrs_post_init__(self):
        if self.until is None:
            raise ValueError("until cannot be None if stopping_condition is None")
        if self.until is not None and self.until <= self.when:
            raise ValueError("until must be > when")

    def is_fixed(self, pop: fwdpy11.DiploidPopulation, key: tuple) -> bool:
        for f in pop.fixations:
            if f.key == key:
                return True
        return False

    def __call__(self, pop, index: int, key: tuple) -> SimulationStatus:
        if pop.generation == self.until:
            if pop.mutations[index].key != key and self.is_fixed(pop, key) is False:
                return SimulationStatus.Restart
            if pop.mcounts[index] > 0:
                return SimulationStatus.Success
            if self.is_fixed(pop, key):
                return SimulationStatus.Success
        return SimulationStatus.Continue


@attr.s(auto_attribs=True)
class _Recorder:
    when: typing.Optional[int]
    until: typing.Optional[int]
    criterion: typing.Callable[
        [fwdpy11.DiploidPopulation, int, typing.Tuple[float, float, int]],
        SimulationStatus,
    ]
    monitor: _ProgressMonitor
    sampling_policy: AncientSamplePolicy

    def __attrs_post_init__(self):
        if self.sampling_policy != AncientSamplePolicy.NEVER:
            if self.when is None or self.when < 0:
                raise ValueError(
                    f"sampling policy is {self.sampling_policy} and when is {self.when}"
                )

    def __call__(self, pop, sampler) -> None:
        if self.monitor.status != SimulationStatus.Success:
            self.monitor.status = self.criterion(
                pop, self.monitor.index, self.monitor.key
            )
        else:
            return

        x = (
            self.sampling_policy == AncientSamplePolicy.DURATION
            and pop.generation >= self.when
            and (self.until is None or pop.generation <= self.until)
        )

        y = (
            self.sampling_policy == AncientSamplePolicy.COMPLETION
            and self.monitor.status == SimulationStatus.Success
            and (self.until is None or pop.generation == self.until)
        )

        if x is True or y is True:
            # Record all alive individuals
            sampler.assign(np.arange(pop.N, dtype=np.uint32))


def _integer_count_details(
    minimum: int,
    maximum: int,
    deme: typing.Optional[int],
    pop: fwdpy11.DiploidPopulation,
):
    if deme is None:
        for c in [minimum, maximum]:
            if c >= 2 * pop.N:
                raise ValueError(
                    f"count {c} "
                    f"is not compatible with total population size of {pop.N}"
                )
    else:
        deme_sizes = pop.deme_sizes(as_dict=True)
        if deme not in deme_sizes:
            raise ValueError(
                f"deme {deme} not present "
                f"at time {pop.generation}. Deme sizes are {deme_sizes}"
            )
        for c in [minimum, maximum]:
            if c >= 2 * pop.N:
                raise ValueError(
                    f"count {c} "
                    f"is not compatible with total population size of {pop.N}"
                )
    return range(minimum, maximum + 1)


def _get_allele_count_range(
    pop: fwdpy11.DiploidPopulation, mutation_parameters: NewMutationParameters
):
    if isinstance(mutation_parameters.frequency, AlleleCount):
        return _integer_count_details(
            mutation_parameters.frequency.count,
            mutation_parameters.frequency.count,
            mutation_parameters.deme,
            pop,
        )
    elif isinstance(mutation_parameters.frequency, AlleleCountRange):
        return _integer_count_details(
            mutation_parameters.frequency.minimum,
            mutation_parameters.frequency.maximum,
            mutation_parameters.deme,
            pop,
        )
    elif isinstance(mutation_parameters.frequency, FrequencyRange):
        if mutation_parameters.deme is None:
            lo = int(np.ceil(mutation_parameters.frequency.minimum * pop.N))
            hi = int(np.floor(mutation_parameters.frequency.minimum * pop.N))
        else:
            deme_sizes = pop.deme_sizes(as_dict=True)
            if mutation_parameters.deme not in deme_sizes:
                raise ValueError(
                    f"deme {mutation_parameters.deme} not present "
                    "at time {pop.generation}"
                )
            lo = int(
                np.ceil(
                    mutation_parameters.frequency.minimum
                    * deme_sizes[mutation_parameters.deme]
                )
            )
            hi = int(
                np.floor(
                    mutation_parameters.frequency.minimum
                    * deme_sizes[mutation_parameters.deme]
                )
            )
        return _integer_count_details(lo, hi, mutation_parameters.deme, pop)
    else:
        raise TypeError(f"unsupported type {type(mutation_parameters.frequency)}")


def _copy_pop_and_add_mutation(
    rng,
    pop,
    params,
    mutation_parameters,
    when,
    evolvets_options,
    sampling_policy,
):
    idx = None
    final_count = None
    pcopy = None
    out_params = None
    if when is None or when == 0:
        count_range = _get_allele_count_range(pop, mutation_parameters)

        for c in count_range:
            pcopy = copy.deepcopy(pop)
            _mutation_params = {
                "window": (
                    mutation_parameters.position.left,
                    mutation_parameters.position.right,
                ),
                "ndescendants": c,
                "data": mutation_parameters.data,
                "deme": mutation_parameters.deme,
            }
            idx = pcopy.add_mutation(rng, **_mutation_params)
            if idx is not None:
                final_count = c
                break
        if idx is None:
            raise AddMutationFailure("failed to add mutation")
        out_params = copy.deepcopy(params)
    else:
        if when < 0:
            raise ValueError(f"when must be >= 0, got {when}")

        pcopy = copy.deepcopy(pop)
        pre_sweep_pdict = {k: copy.deepcopy(v) for k, v in params.asdict().items()}
        pre_sweep_pdict["simlen"] = when
        pre_sweep_params = fwdpy11.ModelParams(**pre_sweep_pdict)
        fwdpy11.evolvets(rng, pcopy, pre_sweep_params, **evolvets_options.asdict())

        count_range = _get_allele_count_range(pcopy, mutation_parameters)
        for c in count_range:
            _mutation_params = {
                "window": (
                    mutation_parameters.position.left,
                    mutation_parameters.position.right,
                ),
                "ndescendants": c,
                "data": mutation_parameters.data,
                "deme": mutation_parameters.deme,
            }
            idx = pcopy.add_mutation(rng, **_mutation_params)
            if idx is not None:
                final_count = c
                break

        if idx is None:
            raise AddMutationFailure("failed to add mutation")
        sweep_pdict = {k: v for k, v in pre_sweep_params.asdict().items()}
        sweep_pdict["simlen"] = params.simlen - pcopy.generation
        out_params = fwdpy11.ModelParams(**sweep_pdict)

    return pcopy, idx, final_count, out_params


def _track_added_mutation(
    rng: fwdpy11.GSLrng,
    pop: fwdpy11._types.DiploidPopulation,
    params: fwdpy11.ModelParams,
    mutation_parameters: NewMutationParameters,
    *,
    when: typing.Optional[int] = None,
    until: typing.Optional[int] = None,
    max_attempts: typing.Optional[int] = None,
    sampling_policy: typing.Optional[AncientSamplePolicy] = None,
    stopping_condition=None,
    evolvets_options: typing.Optional[EvolveOptions] = None,
    return_when_stopping_condition_met: bool = False,
) -> ConditionalModelOutput:
    """
    Track the fate of a specific mutation added to the population.

    :param rng: A random number generator
    :type rng: :class:`fwdpy11.GSLrng`
    :param pop: The input population
    :type pop: :class:`fwdpy11.DiploidPopulation`
    :param params: Input simulation parameters
    :type params: :class:`fwdpy11.ModelParams`
    :param mutation_parameters: Details of the new mutation
    :type mutation_parameters: :class:`NewMutationParameters`
    :param when: Time to add the new mutation
    :type when: int
    :param until: Time stop monitoring the new mutation
    :type until: int
    :param max_attempts: Maximum number of attempts to
                         satisfy `stopping_condition`
    :type max_attempts: int
    :param sampling_policy: Enumeration specifying the
                            recording of ancient samples.
    :type sampling_policy: :class:`AncientSamplePolicy`
    :param stopping_condition: An optional callable that specifies a TODO
    :type stopping_condition: typing.Optional[typing.Callable[
                              [fwdpy11.DiploidPopulation,
                              int,
                              typing.Tuple[float, float, int]],
                              SimulationStatus,]
    :param evolvets_options: Options to :func:`fwdpy11.evolvets`
    :type evolvets_options: :class:`EvolveOptions`
    :param return_when_stopping_condition_met: If `True`,
        return to calling environment once `stopping_condition` is satisfied.
        If `False`, simulate to the end of the model.
    """

    if pop.generation > 0:
        raise ValueError(
            f"the population has been evolved: genration = {pop.generation}"
        )

    if max_attempts is not None:
        if max_attempts < 1:
            raise ValueError(f"max_attempts must be > 0, got {max_attempts}")

    # TODO: all of the stuff below is a hot mess,
    # which would be fixed by the above todo
    # re: arguments to this fn.
    if evolvets_options is None:
        _evolvets_options = EvolveOptions()
    else:
        _evolvets_options = copy.deepcopy(evolvets_options)

    pcopy, idx, ndescendants, local_params = _copy_pop_and_add_mutation(
        rng, pop, params, mutation_parameters, when, _evolvets_options, sampling_policy
    )

    if idx is None:
        raise AddMutationFailure()

    if sampling_policy is None:
        _sampling_policy = AncientSamplePolicy.NEVER
    else:
        _sampling_policy = sampling_policy

    if stopping_condition is None:
        _stopping_condition = _MutationPresent(when, until)  # type: ignore
    else:
        _stopping_condition = stopping_condition

    recorder = _Recorder(
        when,
        until,
        _stopping_condition,
        _ProgressMonitor(
            idx,
            pcopy.mutations[idx].key,
            return_when_stopping_condition_met,
            SimulationStatus.Continue,
        ),
        _sampling_policy,
    )

    tracked_mutation_key = pcopy.mutations[idx].key

    internal_options = _InternalSweepEvolveOptions()

    finished = False

    # Rejection-sample our way to glory
    pop_to_return = None

    attempt = 0
    while (
        finished is not True
        and max_attempts is None
        or max_attempts is not None
        and attempt < max_attempts
    ):
        # NOTE: deepcopy and not copy!
        pcopy_loop = copy.deepcopy(pcopy)
        local_params_copy = copy.deepcopy(local_params)
        evolvets_options_copy = copy.deepcopy(_evolvets_options)

        # evolve
        assert evolvets_options_copy is not None
        fwdpy11.evolvets(
            rng,
            pcopy_loop,
            local_params_copy,
            recorder=recorder,
            stopping_criterion=recorder.monitor,
            **evolvets_options_copy.asdict(),  # type: ignore
            **internal_options.asdict(),  # type: ignore
        )

        # The sim ended, so
        # check if condition was satisfied or not
        if recorder.monitor.status == SimulationStatus.Success:
            pop_to_return = pcopy_loop
            _evolvets_options = evolvets_options_copy
            local_params = local_params_copy
            finished = True

        attempt += 1

    if attempt == max_attempts:
        raise OutOfAttempts()

    assert pop_to_return is not None

    if idx < len(pop_to_return.mutations):
        mutation_index = idx
    else:
        mutation_index = None
    if mutation_index is not None:
        assert pop_to_return.mutations[mutation_index].key == tracked_mutation_key
    fixation_index = None
    for i, f in enumerate(pop_to_return.fixations):
        if f.key == tracked_mutation_key:
            fixation_index = i
    return ConditionalModelOutput(
        pop=pop_to_return,
        params=local_params,
        mutation_index=mutation_index,
        fixation_index=fixation_index,
        num_descendant_nodes=ndescendants,
    )



import copy
import unittest

import numpy as np

import fwdpy11
from test_tree_sequences import set_up_quant_trait_model, set_up_standard_pop_gen_model


def _count_mutations_from_diploids(pop):
    mc = np.zeros(len(pop.mutations), dtype=np.uint32)
    for d in pop.diploids:
        for k in pop.haploid_genomes[d.first].mutations:
            mc[k] += 1
        for k in pop.haploid_genomes[d.first].smutations:
            mc[k] += 1
        for k in pop.haploid_genomes[d.second].mutations:
            mc[k] += 1
        for k in pop.haploid_genomes[d.second].smutations:
            mc[k] += 1
    return mc


def _compare_counts_for_nonneutral_variants(pop, mc):
    w = np.array(
        [i for i, j in enumerate(pop.mutations) if j.neutral is False], dtype=np.uint32
    )
    return all(pop.mcounts[w] == mc[w]) is True


class TestKeepFixations(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.params, self.rng, self.pop = set_up_quant_trait_model()
        pdict = self.params.asdict()
        pdict["rates"] = (
            1e-3,
            self.params.rates.selected_mutation_rate,
            self.params.rates.recombination_rate,
        )
        pdict["nregions"] = [fwdpy11.Region(0, 1, 1)]
        self.params = fwdpy11.ModelParams(**pdict)

    def test_mutation_counts_with_indexing(self):
        """
        Tests atypical use case where neutral mutations are placed into genomes
        """
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        mc = _count_mutations_from_diploids(self.pop)
        self.assertTrue(_compare_counts_for_nonneutral_variants(self.pop, mc))

    def test_mutation_counts_with_indexing_suppressed(self):
        """
        Tests atypical use case where neutral mutations are placed into genomes
        """
        fwdpy11.evolvets(
            self.rng, self.pop, self.params, 100, suppress_table_indexing=True
        )
        mc = _count_mutations_from_diploids(self.pop)
        self.assertTrue(_compare_counts_for_nonneutral_variants(self.pop, mc))

    def test_mutation_counts_with_indexing_suppressed_no_neutral_muts_in_genomes(self):
        """
        A sim w/ and w/o putting neutral variants in genomes
        should give the same mutation counts.
        """
        pop2 = copy.deepcopy(self.pop)
        rng = fwdpy11.GSLrng(101 * 45 * 110 * 210)  # Use same seed!!!
        pdict = self.params.asdict()
        pdict["prune_selected"] = False
        self.params = fwdpy11.ModelParams(**pdict)
        params = copy.deepcopy(self.params)
        fwdpy11.evolvets(rng, pop2, params, 100, suppress_table_indexing=True)
        fwdpy11.evolvets(
            self.rng, self.pop, self.params, 100, suppress_table_indexing=True
        )
        ti = fwdpy11.TreeIterator(
            self.pop.tables, [i for i in range(2 * self.pop.N)])
        mc = _count_mutations_from_diploids(self.pop)
        # This assert was added when working
        # on GitHub isuse 954
        for m in pop2.tables.mutations:
            assert m.key < len(pop2.tables.mutations)
        for t in ti:
            for m in t.mutations():
                # This assert was added when working
                # on GitHub isuse 954
                assert m.key < len(pop2.mutations)
                # Have to skip neutral mutations b/c they won't
                # end up in mc b/c it is obtained from genomes
                if pop2.mutations[m.key].neutral is False:
                    self.assertEqual(mc[m.key], self.pop.mcounts[m.key])
                self.assertEqual(t.leaf_counts(m.node), pop2.mcounts[m.key])


class TestPruneFixations(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.params, self.rng, self.pop = set_up_standard_pop_gen_model()
        pdict = self.params.asdict()
        pdict["rates"] = (
            1e-3,
            self.params.rates.selected_mutation_rate,
            self.params.rates.recombination_rate,
        )
        pdict["nregions"] = [fwdpy11.Region(0, 1, 1)]
        self.params = fwdpy11.ModelParams(**pdict)

    @unittest.skip("doesn't test anything...")
    def test_no_mutations(self):
        pdict = self.params.asdict()
        pdict["rates"] = (0, 0, 0)
        self.params = fwdpy11.ModelParams(**pdict)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)

    def test_mutation_counts_with_indexing(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        mc = _count_mutations_from_diploids(self.pop)
        self.assertTrue(_compare_counts_for_nonneutral_variants(self.pop, mc))
        self.assertEqual(len(np.where(mc == 2 * self.pop.N)[0]), 0)

    def test_mutation_counts_with_indexing_suppressed(self):
        fwdpy11.evolvets(
            self.rng, self.pop, self.params, 100, suppress_table_indexing=True
        )
        mc = _count_mutations_from_diploids(self.pop)
        self.assertTrue(_compare_counts_for_nonneutral_variants(self.pop, mc))
        self.assertEqual(len(np.where(mc == 2 * self.pop.N)[0]), 0)


class TestNeutralMutRegions(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.params, self.rng, self.pop = set_up_standard_pop_gen_model()
        pdict = self.params.asdict()
        pdict["rates"] = (
            1e-3,
            self.params.rates.selected_mutation_rate,
            self.params.rates.recombination_rate,
        )
        pdict["nregions"] = [fwdpy11.Region(
            0, 0.25, 1), fwdpy11.Region(0.5, 1, 1)]
        self.params = fwdpy11.ModelParams(**pdict)

    def test_neutral_mut_locations(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        pos = [
            self.pop.tables.sites[i.site].position
            for i in self.pop.tables.mutations
            if i.neutral
        ]
        self.assertTrue(all([i < 0.25 or i >= 0.5 for i in pos]))


def test_trigger_final_simplification_with_fixations_to_remove():
    """
    MRE that triggers an exception that would hit after many of the
    changes in PR647 were in place.  Commit 06a8a41 avoids the exception.
    """
    burnin = 10
    N = 133
    pdict = {
        # Add a region for neutral mutations:
        "nregions": [fwdpy11.Region(0, 1, 1)],
        "sregions": [fwdpy11.ExpS(beg=0, end=1, weight=1, mean=0.2)],
        "recregions": [fwdpy11.PoissonInterval(0, 1, 1e-2)],
        "gvalue": [fwdpy11.Multiplicative(2.0)],
        "rates": (1e-3, 1e-3, None),
        "simlen": burnin * N,
        "demography": fwdpy11.ForwardDemesGraph.tubes([N],
                                                      burnin=10*N,
                                                      burnin_is_exact=True),
        "prune_selected": True,
    }
    params = fwdpy11.ModelParams(**pdict)
    pop = fwdpy11.DiploidPopulation(N, 1.0)
    rng = fwdpy11.GSLrng(54321)

    fwdpy11.evolvets(
        rng,
        pop,
        params,
        37,
        suppress_table_indexing=False,
    )


if __name__ == "__main__":
    unittest.main()



#
# Copyright(C) 2017 Kevin Thornton < krthornt @uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software : you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.If not, see < http: //www.gnu.org/licenses/>.
#

from typing import Callable, Optional
import warnings

import fwdpy11

from ._fwdpy11 import GSLrng, SampleRecorder
from ._types import DiploidPopulation, ModelParams


def evolvets(
    rng: GSLrng,
    pop: DiploidPopulation,
    params: ModelParams,
    simplification_interval: int,
    recorder: Optional[Callable[[
        DiploidPopulation, SampleRecorder], None]] = None,
    *,
    post_simplification_recorder: Optional[Callable[[
        DiploidPopulation], None]] = None,
    suppress_table_indexing: Optional[bool] = None,
    record_gvalue_matrix: bool = False,
    stopping_criterion: Optional[Callable[[
        DiploidPopulation, bool], bool]] = None,
    track_mutation_counts: bool = False,
    remove_extinct_variants: bool = True,
    preserve_first_generation: bool = False,
):
    """
    Evolve a population with tree sequence recording

    :param rng: random number generator
    :type rng: :class:`fwdpy11.GSLrng`
    :param pop: A population
    :type pop: :class:`fwdpy11.DiploidPopulation`
    :param params: simulation parameters
    :type params: :class:`fwdpy11.ModelParams`
    :param simplification_interval: Number of generations
        between simplifications.
    :type simplification_interval: int
    :param recorder: (None) A temporal sampler/data recorder.
    :type recorder: typing.Callable
    :param post_simplification_recorder: (None) A temporal sampler
    :type post_simplification_recorder: typing.Callable
    :param suppress_table_indexing: (`None`) Prevents edge table indexing until
                                    end of simulation.  The default value
                                    (`None`) will be interpreted as `True`
    :type suppress_table_indexing: Optional[bool]
    :param record_gvalue_matrix: (False) Whether to record genetic values into
                                 :attr:`fwdpy11.DiploidPopulation.genetic_values`.
    :type record_gvalue_matrix: bool
    :param preserve_first_generation: (False) Whether to record generation 0 as
                                      ancient samples. Must be `True` for
                                      tree sequence "recapitation". See
                                      :ref:`recapitation`.
    :type preserve_first_generation: bool

    The recording of genetic values into :attr:`fwdpy11.DiploidPopulation.genetic_values`
    is suppressed by default.  First, it is redundant with
    :attr:`fwdpy11.DiploidMetadata.g` for the common case of mutational effects on a
    single trait. Second, we save some memory by not tracking these matrices.
    However, it is useful to track these data for some
    cases when simulating multivariate mutational effects (pleiotropy).

    .. note::
        If recorder is None,
        then :class:`fwdpy11.NoAncientSamples` will be used.

        If post_simplification_recorder is None, then
        :class:`fwdpy11.RecordNothing` will be used.

    .. versionchanged:: 0.5.2

        Added post_simplification_recorder.

    .. versionchanged:: 0.7.1

        Added preserve_first_generation.

    .. versionchanged:: 0.8.0

        Update to refactored ModelParams.
        Added ``check_demographic_event_timings``.

    .. versionchanged: 0.18.0

        `suppress_table_indexing` default changed to `None`,
        which is interpreted as `True`.

    .. versionchanged: 0.20.0

        * Back-end refactored to use demes models directly.
        * Will define a default demographic model using
          :func:`fwdpy11.DemesForwardGraph.tubes` if the
          input model is `None`.
        * Remove option `check_demographic_event_timings`

    """
    if params.demography is not None:
        try:
            # DemographicModelDetails ?
            demographic_model = params.demography.model
        except AttributeError:
            # No, assume it is a DiscreteDemography
            demographic_model = params.demography
        except Exception as e:
            raise e
    else:
        # Build a default model of "tubes"
        from fwdpy11 import ForwardDemesGraph
        sizes = pop.deme_sizes()[1].tolist()
        msg = "Applying a default demographic model "
        msg += f"where deme sizes are {sizes} "
        msg += f"and the burn-in length is 10*{sum(sizes)}. "
        msg += "This will raise an error in future releases."
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        demographic_model = ForwardDemesGraph.tubes(sizes, 10)

    for r in params.sregions:
        if isinstance(r, fwdpy11.mvDES):
            if isinstance(r.des, list):
                nd = demographic_model.number_of_demes()
                if nd > params.gvalue.ndemes:
                    msg = f"maxmimum number of demes in model ({nd})"
                    msg += " is not compatible with genetic value"
                    msg += f" number of demes {params.gvalue.ndemes}"
                    raise ValueError(msg)
                for rr in r.des:
                    assert isinstance(rr, fwdpy11.Sregion)  # type: ignore
                    if rr.beg < 0:
                        raise ValueError(f"{r} has begin value < 0.0")
                    if rr.end > pop.tables.genome_length:
                        raise ValueError(
                            f"{r} has end value >= genome length"
                            " of {pop.tables.genome_length}"
                        )
            else:
                assert isinstance(r, fwdpy11.mvDES), f"{type(r)}"
                assert isinstance(r.des, fwdpy11.Sregion)  # type: ignore
                nd = demographic_model.number_of_demes()
                if isinstance(params.gvalue, list):
                    for g in params.gvalue:
                        if nd > g.ndemes:
                            msg = f"maxmimum number of demes in model ({nd})"
                            msg += " is not compatible with genetic value"
                            msg += f" number of demes {params.gvalue.ndemes}"
                            raise ValueError(msg)
                else:
                    if nd > params.gvalue.ndemes:
                        msg = f"maxmimum number of demes in model ({nd})"
                        msg += " is not compatible with genetic value"
                        msg += f" number of demes {params.gvalue.ndemes}"
                        raise ValueError(msg)
                if r.des.beg < 0:
                    raise ValueError(f"{r} has begin value < 0.0")
                if r.des.end > pop.tables.genome_length:
                    raise ValueError(
                        f"{r} has end value >= genome"
                        " length of {pop.tables.genome_length}"
                    )
        else:
            assert isinstance(r, fwdpy11.Sregion)  # type: ignore
            if r.beg < 0:
                raise ValueError(f"{r} has begin value < 0.0")
            if r.end > pop.tables.genome_length:
                raise ValueError(
                    f"{r} has end value >= genome"
                    " length of {pop.tables.genome_length}"
                )

    for r in params.nregions:
        assert isinstance(r, fwdpy11.Region)
        if r.beg < 0:
            raise ValueError(f"{r} has begin value < 0.0")
        if r.end > pop.tables.genome_length:
            raise ValueError(
                f"{r} has end value > genome length of {pop.tables.genome_length}"
            )

    for r in params.recregions:
        assert isinstance(r, fwdpy11._fwdpy11.PoissonCrossoverGenerator) or \
            isinstance(
                r, fwdpy11._fwdpy11.NonPoissonCrossoverGenerator) or \
            isinstance(r, fwdpy11.Region), f"{type(r)}"
        if hasattr(r, "beg") and hasattr(r, "end"):
            if r.beg < 0:
                raise ValueError(f"{r} has begin value < 0.0")
            if r.end > pop.tables.genome_length:
                raise ValueError(
                    f"{r} has end value >= genome length of {pop.tables.genome_length}"
                )
        elif hasattr(r, "position"):
            if r.position < 0:
                raise ValueError(f"{r} has position value < 0.0")
            if r.position >= pop.tables.genome_length:
                raise ValueError(
                    f"{r} has position value >= genome length of {pop.tables.genome_length}"
                )
        elif hasattr(r, "regions"):
            for i in r.regions:
                if i.beg < 0:
                    raise ValueError(f"{r} has begin value < 0.0")
                if i.end > pop.tables.genome_length:
                    raise ValueError(
                        f"{r} has end value >= genome length of {pop.tables.genome_length}"
                    )
        else:
            raise TypeError(f"unexpceted type: {type(r)}")

    if recorder is None:
        from ._fwdpy11 import NoAncientSamples

        recorder = NoAncientSamples()

    if post_simplification_recorder is None:
        from ._fwdpy11 import RecordNothing

        post_simplification_recorder = RecordNothing()
        reset_treeseqs_after_simplify = False
    else:
        reset_treeseqs_after_simplify = True

    if stopping_criterion is None:
        from ._fwdpy11 import _no_stopping

        stopping_criterion = _no_stopping

    from ._fwdpy11 import (
        MutationRegions,
        dispatch_create_GeneticMap,
        dispatch_create_GeneticMap_non_Region,
        evolve_with_tree_sequences,
    )

    fwdpy11._validate_regions(params.sregions, pop.tables.genome_length)
    fwdpy11._validate_regions(params.nregions, pop.tables.genome_length)

    pneutral = 0.0
    if params.rates.neutral_mutation_rate + params.rates.selected_mutation_rate > 0.0:
        pneutral = params.rates.neutral_mutation_rate / (
            params.rates.neutral_mutation_rate + params.rates.selected_mutation_rate
        )
    mm = MutationRegions.create(pneutral, params.nregions, params.sregions)

    if all([isinstance(i, fwdpy11.Region) for i in params.recregions]):
        rm = dispatch_create_GeneticMap(
            params.rates.recombination_rate, params.recregions)
    else:
        poisson = [i for i in params.recregions if isinstance(
            i, fwdpy11._fwdpy11.PoissonCrossoverGenerator)]
        non_poisson = [i for i in params.recregions if isinstance(
            i, fwdpy11._fwdpy11.NonPoissonCrossoverGenerator)]
        rm = dispatch_create_GeneticMap_non_Region(poisson, non_poisson)
        assert rm._num_poisson_callbacks() == len(poisson)
        assert rm._num_non_poisson_callbacks() == len(non_poisson)

    from ._fwdpy11 import SampleRecorder, _evolve_with_tree_sequences_options

    sr = SampleRecorder()
    from ._fwdpy11 import _dgvalue_pointer_vector

    options = _evolve_with_tree_sequences_options()

    options.preserve_selected_fixations = params.prune_selected is False
    if suppress_table_indexing is not None:
        options.suppress_edge_table_indexing = suppress_table_indexing
    else:
        options.suppress_edge_table_indexing = True
    options.record_gvalue_matrix = record_gvalue_matrix
    options.track_mutation_counts_during_sim = track_mutation_counts
    options.remove_extinct_mutations_at_finish = remove_extinct_variants
    options.reset_treeseqs_to_alive_nodes_after_simplification = \
        reset_treeseqs_after_simplify
    options.preserve_first_generation = preserve_first_generation
    options.allow_residual_selfing = params.allow_residual_selfing

    if options.allow_residual_selfing is False:
        from fwdpy11._types.forward_demes_graph import _round_via_decimal
        for d, deme in enumerate(demographic_model.graph.demes):
            for e, epoch in enumerate(deme.epochs):
                start_size = _round_via_decimal(epoch.start_size)
                end_size = _round_via_decimal(epoch.end_size)
                if start_size == 1 or end_size == 1:
                    msg = f"deme {d} epoch {e} has a size of 1"
                    msg += " but residual selfing is not allowed."
                    msg += " This model will raise an error if this deme "
                    msg += "contributes to ancestry after the size of 1"
                    msg += " is reached."
                    warnings.warn(msg, UserWarning, stacklevel=2)

    gvpointers = _dgvalue_pointer_vector(params.gvalue)
    evolve_with_tree_sequences(
        rng,
        pop,
        sr,
        simplification_interval,
        demographic_model,
        params.simlen,
        params.rates.neutral_mutation_rate,
        params.rates.selected_mutation_rate,
        mm,
        rm,
        gvpointers,
        recorder,
        stopping_criterion,
        post_simplification_recorder,
        options,
    )



#
# Copyright (C) 2022 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import msprime
import numpy as np
import pytest

from hypothesis import given
from hypothesis import settings
from hypothesis.strategies import integers

import fwdpy11


def make_position(sites, edge):
    position = float(np.random.uniform(edge.left, edge.right, 1)[0])
    while position in sites:
        position = float(np.random.uniform(edge.left, edge.right, 1)[0])
    sites.add(position)
    return position, sites


def generate_mutation_origin_time(tables, edge):
    origin_time = int(tables.nodes.time[edge.child] +
                      (tables.nodes.time[edge.parent] -
                       tables.nodes.time[edge.child])/2.0)
    # Sometimes a branch is too short to assign an integer
    # time, so we just bail
    if origin_time >= tables.nodes.time[edge.parent] or \
            origin_time < tables.nodes.time[edge.child]:
        return None

    assert origin_time >= tables.nodes.time[edge.child]
    assert origin_time < tables.nodes.time[edge.parent]
    return origin_time


def make_generic_mutation_metadata(origin_time):
    return {'s': 1e-3, 'h': 1.0, 'origin':
            origin_time,
            'label': 0,
            'neutral': 0,
            # fwdpy11 will have to figure out a real key value later.
            'key': 0}


def make_treeseq_with_one_row_containing_bad_metadata(seed, callback):
    ts = msprime.sim_ancestry(5, population_size=10000, random_seed=seed)

    tables = ts.tables

    tables.mutations.metadata_schema = \
        fwdpy11.tskit_tools.metadata_schema.MutationMetadata
    sites = set()
    for edge in tables.edges:
        position, sites = make_position(sites, edge)
        site = tables.sites.add_row(position, '0')
        origin_time = generate_mutation_origin_time(tables, edge)
        if origin_time is not None:
            md = make_generic_mutation_metadata(origin_time)
            callback(md)
            tables.mutations.add_row(
                site, edge.child, '1', time=origin_time, metadata=md)
            sites.add(position)
            break

    assert tables.mutations.num_rows > 0

    tables.sort()

    ts = tables.tree_sequence()

    return ts

# Failures:
#           recrate=0.00010693073272705097, seed2=1, seed=1,


@given(integers(1, int(2**32 - 1)), integers(1, int(2**32 - 1)))
@settings(deadline=None, max_examples=10)
def test_import_msprime_mutations(seed, seed2):
    ts = msprime.sim_ancestry(5, population_size=10000,
                              recombination_rate=1e-5, sequence_length=1.0,
                              discrete_genome=False, random_seed=seed)
    np.random.seed(seed2)
    tables = ts.tables
    tables.mutations.metadata_schema = \
        fwdpy11.tskit_tools.metadata_schema.MutationMetadata
    sites = set()
    for edge in tables.edges:
        position, sites = make_position(sites, edge)
        origin_time = generate_mutation_origin_time(tables, edge)
        if origin_time is not None:
            md = make_generic_mutation_metadata(origin_time)
            site = tables.sites.add_row(position, '0')
            tables.mutations.add_row(
                site, edge.child, '1', time=origin_time, metadata=md)

    tables.sort()

    ts = tables.tree_sequence()
    pop = fwdpy11.DiploidPopulation.create_from_tskit(
        ts, import_mutations=True)

    assert len(pop.mutations) == ts.num_mutations
    assert len(pop.tables.sites) == ts.num_sites

    for site in ts.sites():
        assert len([i for i in pop.mutations if i.pos == site.position]) == 1
        assert len(
            [i for i in pop.tables.sites if i.position == site.position]) == 1
        assert len(
            [i for i in pop.tables.mutations if
             pop.tables.sites[i.key].position == site.position]) == 1

    # All origin times must be < the time of the mutation time
    for m in pop.tables.mutations:
        node_time = pop.tables.nodes[m.node].time
        assert pop.mutations[m.key].g <= node_time

    # manually tally mutation counts and compare to genomes
    counts = {}
    for g in pop.haploid_genomes:
        for m in g.smutations.tolist() + g.mutations.tolist():
            if m in counts:
                counts[m] += 1
            else:
                counts[m] = 1
    for i, j in counts.items():
        assert pop.mcounts[i] == j

    assert np.array(pop.diploids).flatten().tolist() == [
        (2*i, 2*i+1) for i in range(pop.N)]

    # Finally make sure that which samples have mutations
    # matches up with the tskit input
    for t in ts.trees(sample_lists=True):
        for m in t.mutations():
            samples = [i for i in t.samples(m.node)]
            count = len(samples)
            for i, mut in enumerate(pop.mutations):
                if mut.pos == ts.site(m.site).position:
                    assert count == pop.mcounts[i]

                # now, we brute force get a sample
                # list from our pop w/o traversing the trees
                fp11samples = []
                for dip, md in zip(pop.diploids, pop.diploid_metadata):
                    for key in pop.haploid_genomes[dip.first].smutations:
                        if pop.mutations[key].pos == ts.site(m.site).position:
                            fp11samples.append(md.nodes[0])
                    for key in pop.haploid_genomes[dip.second].smutations:
                        if pop.mutations[key].pos == ts.site(m.site).position:
                            fp11samples.append(md.nodes[1])
                samples = sorted(samples)
                fp11samples = sorted(fp11samples)
                assert samples == fp11samples


@given(integers(1, int(2**32 - 1)), integers(1, int(2**32 - 1)))
@settings(deadline=None, max_examples=10)
def test_import_msprime_mutations_bad_metadata(seed1, seed2):
    np.random.seed(seed1)

    def bad_effect_size(md):
        md['s'] = np.nan

    def bad_dominance(md):
        md['h'] = np.nan

    # The "origin time" in the metadata
    # must equal the mutation's time value

    # NOTE: disabled in 0.19.9 to allow
    # creating of pops from fwdpy11-generated
    # mutations
    # def bad_origin(md):
    #     if md['origin'] > 0.0:
    #         md['origin'] = -1*md['origin']
    #     else:
    #         md['origin'] = -1

    # NOTE: disable in 0.19.8 to
    # fix #1109 but this should be
    # a requirement for mutations added "manually"
    # to tskit.
    # def bad_origin2(md):
    #     if md['origin'] > 0.0:
    #         md['origin'] = -1 #int(0.5*md['origin'])
    #     else:
    #         md['origin'] = -1

    # This will not trigger an error:
    # fwdpy11 sets the neutral field based on if effect_size == 0.0
    # def neutrality_mismatch1(md):
    #     md['neutral'] = 1

    # NOTE: bring back bad_origin and bad_origin2 in 0.20.0 if possible
    for callback in [bad_effect_size, bad_dominance]:
        ts = make_treeseq_with_one_row_containing_bad_metadata(seed2, callback)
        with pytest.raises(ValueError):
            _ = fwdpy11.DiploidPopulation.create_from_tskit(
                ts, import_mutations=True)



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from fwdpy11._types.forward_demes_graph import ForwardDemesGraph
from fwdpy11._types.demographic_model_details import DemographicModelDetails
from .diploid_population import DiploidPopulation


@dataclass
class DemographyDebugger(object):
    """
    Efficiently debug demographic events.

    The class executes a mock simulation
    of demographic events in order to quickly
    detect errors. It also generates a "report"
    of the model details for double-checking.

    If model errors are found, a ``ValueError`` exception
    is raised. The goal is to catch the types of errors that would
    result in an exception during a simulation.

    The class is initialized with the following arguments,
    which may be either positional or keyword:

    :param initial_deme_sizes: The initial sizes of each deme
    :type initial_deme_sizes: list or fwdpy11.DiploidPopulation
    :param events: The demographic model
    :type events: fwdpy11.ForwardDemesGraph or
                  fwdpy11.DemographicModelDetails
    :param simlen: The length of the simulation.  Defaults to `None`.
    :type simlen: int
    :param deme_labels: A map from deme index to a printable name
    :type deme_labels: dict

    .. versionadded:: 0.6.0

    .. versionchanged:: 0.8.1

        Initialization now done with attr.
        A list of initial deme sizes is now accepted.
    """

    initial_deme_sizes: Union[List[int], Dict[int, int], DiploidPopulation]
    events: Union[ForwardDemesGraph, DemographicModelDetails]
    simlen: Optional[int] = None
    deme_labels: Optional[Dict] = None



import demes
import fwdpy11
import gvalue_recorder

N = 1000

yaml = """
time_units: generations
demes:
 - name: pop
   epochs:
    - start_size: 1000
"""

graph = demes.loads(yaml)
demography = fwdpy11.discrete_demography.from_demes(graph, 1)

pdict = {
    "demography": demography,
    "simlen": 100,
    "nregions": [],
    "sregions": [fwdpy11.GaussianS(0, 1, 1, 0.25)],
    "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-3)],
    "rates": (0.0, 5e-3, None),
    "gvalue": fwdpy11.Additive(
        2.0,
        fwdpy11.GaussianStabilizingSelection.single_trait(
            optima=[fwdpy11.Optimum(VS=1.0, optimum=1.0, when=0)]
        ),
    ),
    "prune_selected": False,
}

params = fwdpy11.ModelParams(**pdict)

pop = fwdpy11.DiploidPopulation(N, 1.0)

rng = fwdpy11.GSLrng(42)

gvalues = []
rg = gvalue_recorder.record_gvalue()


def r(pop, _):
    rg(pop, gvalues)


fwdpy11.evolvets(rng, pop, params, 100, r)

assert len(gvalues) == pop.generation, "Callback failure"



import math
import pickle
import unittest

import numpy as np

import fwdpy11 as fp11
import snowdrift


class SamplePhenotypes(object):
    """
    Temporal sampler checks that one can hook
    a stateful fitness model to a sampler
    and access its data and that the data
    are as expected.
    """

    def __init__(self, f, slope, p0):
        self.f = f
        self.slope = slope
        self.p0 = p0
        self.sig0 = (1.0 / slope) * math.log(self.p0 / (1.0 - self.p0))

    def __call__(self, pop, sampler):
        for i in range(pop.N):
            w = 0.0
            for m in pop.haploid_genomes[pop.diploids[i].first].smutations:
                w += pop.mutations[m].s
            for m in pop.haploid_genomes[pop.diploids[i].second].smutations:
                w += pop.mutations[m].s
            w = 1.0 / (1.0 + math.exp(-self.slope * (w + self.sig0)))
            assert np.isclose(w, self.f.phenotypes[i])


def evolve_snowdrift(args):
    """
    We write the function taking a tuple
    out of habit, simplifying later
    integration with multiprocessing or
    concurrent.futures.
    """
    N, seed = args
    # Construct as single-deme object
    # with N diploids
    pop = fp11.DiploidPopulation(N, 1.0)
    # Initialize a random number generator
    rng = fp11.GSLrng(seed)
    p = {
        "sregions": [fp11.ExpS(0, 1, 1, -0.1, 1.0)],
        "recregions": [fp11.PoissonInterval(0, 1, 1e-3)],
        "nregions": [],
        "gvalue": snowdrift.DiploidSnowdrift(0.2, -0.2, 1, -2, 1, 0.3),
        # evolve for 100 generations so that unit tests are
        # fast
        "demography": fp11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                   burnin=20,
                                                   burnin_is_exact=True),
        "simlen": 20,
        "rates": (0.0, 0.0025, None),
        "prune_selected": False,
    }
    params = fp11.ModelParams(**p)
    sampler = SamplePhenotypes(params.gvalue, 1, 0.3)
    fp11.evolvets(rng, pop, params, 100, sampler)
    # return our pop
    return pop


class testSnowdrift(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.f = snowdrift.DiploidSnowdrift(1, -1, 0.1, 0.2, 1, 0.3)

    def testShape(self):
        s = self.f.shape
        self.assertEqual(len(s), 1)
        self.assertEqual(s[0], 1)

    def test_genetic_values(self):
        self.assertEqual(len(self.f.genetic_values), 1)

    def testPickle(self):
        self.f.phenotypes = [1, 2, 3, 4]
        p = pickle.dumps(self.f, -1)
        up = pickle.loads(p)
        self.assertEqual(up.phenotypes, self.f.phenotypes)

    def test_evolve(self):
        p = evolve_snowdrift((1000, 42))
        p


if __name__ == "__main__":
    unittest.main()
